import glob
import itertools
import logging
import random
import re

from abc import ABC, abstractmethod
from enum import Enum
from os import path, sep
from typing import Iterator, List, Tuple, NamedTuple, Optional

from bcipy.helpers.exceptions import BciPyCoreException
from bcipy.helpers.list import grouper

from PIL import Image
# Prevents pillow from filling the console with debug info
logging.getLogger('PIL').setLevel(logging.WARNING)

from psychopy import core
import numpy as np
import sounddevice as sd
import soundfile as sf
from mne import Annotations, Epochs
from mne.io import RawArray
import mne


log = logging.getLogger(__name__)
DEFAULT_FIXATION_PATH = 'bcipy/static/images/main/PLUS.png'
DEFAULT_TEXT_FIXATION = '+'


class StimuliOrder(Enum):
    """Stimuli Order.

    Enum to define the ordering of stimuli for inquiry.
    """
    RANDOM = 'random'
    ALPHABETICAL = 'alphabetical'

    @classmethod
    def list(cls):
        """Returns all enum values as a list"""
        return list(map(lambda c: c.value, cls))


class TargetPositions(Enum):
    """Target Positions.

    Enum to define the positions of targets within the inquiry.
    """
    RANDOM = 'random'
    DISTRIBUTED = 'distributed'

    @classmethod
    def list(cls):
        """Returns all enum values as a list"""
        return list(map(lambda c: c.value, cls))


class PhotoDiodeStimuli(Enum):
    """Photodiode Stimuli.

    Enum to define unicode stimuli needed for testing system timing.
    """

    EMPTY = '\u25A1'  # box with a white border, no fill
    SOLID = '\u25A0'  # solid white box

    @classmethod
    def list(cls):
        """Returns all enum values as a list"""
        return list(map(lambda c: c.value, cls))


class InquirySchedule(NamedTuple):
    """Schedule for the next inquiries to present, where each inquiry specifies
    the stimulus, duration, and color information.

    Attributes
    ----------
    - stimuli: `List[List[str]]`
    - durations: `List[List[float]]`
    - colors: `List[List[str]]`
    """
    stimuli: List[List[str]]
    durations: List[List[float]]
    colors: List[List[str]]


class Reshaper(ABC):

    @abstractmethod
    def __call__(self):
        ...


class InquiryReshaper:
    def __call__(self,
                 trial_targetness_label: List[str],
                 timing_info: List[float],
                 eeg_data: np.ndarray,
                 sample_rate: int,
                 trials_per_inquiry: int,
                 offset: float = 0,
                 channel_map: List[int] = None,
                 poststimulus_length: float = 0.5,
                 prestimulus_length: float = 0.0,
                 transformation_buffer: float = 0.0,
                 target_label: str = 'target') -> Tuple[np.ndarray, np.ndarray]:
        """Extract inquiry data and labels.

        Args:
            trial_targetness_label (List[str]): labels each trial as "target", "non-target", "first_pres_target", etc
            timing_info (List[float]): Timestamp of each event in seconds
            eeg_data (np.ndarray): shape (channels, samples) preprocessed EEG data
            sample_rate (int): sample rate of data provided in eeg_data
            trials_per_inquiry (int): number of trials in each inquiry
            offset (float, optional): Any calculated or hypothesized offsets in timings. Defaults to 0.
            channel_map (List[int], optional): Describes which channels to include or discard.
                 Defaults to None; all channels will be used.
            poststimulus_length (float, optional): time in seconds needed after the last trial in an inquiry.
            prestimulus_length (float, optional): time in seconds needed before the first trial in an inquiry.
            transformation_buffer (float, optional): time in seconds to buffer the end of the inquiry. Defaults to 0.0.
            target_label (str): label of target symbol. Defaults to "target"

        Returns:
            reshaped_data (np.ndarray): inquiry data of shape (Channels, Inquiries, Samples)
            labels (np.ndarray): integer label for each inquiry. With `trials_per_inquiry=K`,
                a label of [0, K-1] indicates the position of `target_label`, or label of K indicates
                `target_label` was not present.
            reshaped_trigger_timing (List[List[int]]): For each inquiry, a list of the sample index where each trial
                begins, accounting for the prestim buffer that may have been added to the front of each inquiry.
        """
        if channel_map:
            # Remove the channels that we are not interested in
            channels_to_remove = [idx for idx, value in enumerate(channel_map) if value == 0]
            eeg_data = np.delete(eeg_data, channels_to_remove, axis=0)

        n_inquiry = len(timing_info) // trials_per_inquiry
        trial_duration_samples = int(poststimulus_length * sample_rate)
        prestimulus_samples = int(prestimulus_length * sample_rate)

        # triggers in seconds are mapped to triggers in number of samples.
        triggers = list(map(lambda x: int((x + offset) * sample_rate), timing_info))

        # First, find the longest inquiry in this experiment
        # We'll add or remove a few samples from all other inquiries, to match this length
        def get_inquiry_len(inq_trigs):
            return inq_trigs[-1] - inq_trigs[0]

        longest_inquiry = max(grouper(triggers, trials_per_inquiry, fillvalue='x'), key=lambda xy: get_inquiry_len(xy))
        num_samples_per_inq = get_inquiry_len(longest_inquiry) + trial_duration_samples
        buffer_samples = int(transformation_buffer * sample_rate)

        # Label for every inquiry
        labels = np.zeros(
            (n_inquiry, trials_per_inquiry), dtype=np.compat.long
        )  # maybe this can be configurable? return either class indexes or labels ('nontarget' etc)
        reshaped_data, reshaped_trigger_timing = [], []
        for inquiry_idx, trials_within_inquiry in enumerate(
            grouper(zip(trial_targetness_label, triggers), trials_per_inquiry, fillvalue='x')
        ):
            first_trigger = trials_within_inquiry[0][1]

            trial_triggers = []
            for trial_idx, (trial_label, trigger) in enumerate(trials_within_inquiry):
                if trial_label == target_label:
                    labels[inquiry_idx, trial_idx] = 1

                # If prestimulus buffer is used, we add it here so that trigger timings will
                # still line up with trial onset
                trial_triggers.append((trigger - first_trigger) + prestimulus_samples)
            reshaped_trigger_timing.append(trial_triggers)
            start = first_trigger - prestimulus_samples
            stop = first_trigger + num_samples_per_inq + buffer_samples
            reshaped_data.append(eeg_data[:, start:stop])

        return np.stack(reshaped_data, 1), labels, reshaped_trigger_timing

    @staticmethod
    def extract_trials(
            inquiries: np.ndarray,
            samples_per_trial: int,
            inquiry_timing: List[List[float]],
            prestimulus_samples: int = 0) -> np.ndarray:
        """Extract Trials.

        After using the InquiryReshaper, it may be necessary to further trial the data for processing.
        Using the number of samples and inquiry timing, the data is reshaped from Channels, Inquiry, Samples to
        Channels, Trials, Samples. These should match with the trials extracted from the TrialReshaper given the same
        slicing parameters.

        Parameters
        ----------
        inquiries : np.ndarray
            shape (Channels, Inquiries, Samples)
        samples_per_trial : int
            number of samples per trial
        inquiry_timing : List[List[float]]
            For each inquiry, a list of the sample index where each trial begins
        prestimulus_samples : int, optional
            Number of samples to move the start of each trial in each inquiry, by default 0.
            This is useful if wanting to use baseline intervals before the trial onset, along with the trial data.

        Returns
        -------
        np.ndarray
            shape (Channels, Trials, Samples)
        """
        new_trials = []
        num_inquiries = inquiries.shape[1]
        for inquiry_idx, timing in zip(range(num_inquiries), inquiry_timing):  # C x I x S

            # time == samples from the start of the inquiry
            for time in timing:
                start = time - prestimulus_samples
                end = time + samples_per_trial

                try:
                    new_trials.append(inquiries[:, inquiry_idx, start:end])
                except IndexError:
                    raise BciPyCoreException(
                        f'InquiryReshaper.extract_trials: index out of bounds. \n'
                        f'Inquiry: [{inquiry_idx}] from {start}:{end}. init_time: {time}, '
                        f'prestimulus_samples: {prestimulus_samples}, samples_per_trial: {samples_per_trial} \n')
        return np.stack(new_trials, 1)  # C x T x S


class TrialReshaper(Reshaper):
    def __call__(self,
                 trial_targetness_label: list,
                 timing_info: list,
                 eeg_data: np.ndarray,
                 sample_rate: int,
                 offset: float = 0,
                 channel_map: List[int] = None,
                 poststimulus_length: float = 0.5,
                 prestimulus_length: float = 0.0,
                 target_label: str = "target") -> Tuple[np.ndarray, np.ndarray]:
        """Extract trial data and labels.

        Parameters
        ----------
            trial_targetness_label (list): labels each trial as "target", "non-target", "first_pres_target", etc
            timing_info (list): Timestamp of each event in seconds
            eeg_data (np.ndarray): shape (channels, samples) preprocessed EEG data
            sample_rate (int): sample rate of preprocessed EEG data
            trials_per_inquiry (int, optional): unused, kept here for consistent interface with `inquiry_reshaper`
            offset (float, optional): Any calculated or hypothesized offsets in timings.
                Defaults to 0.
            channel_map (List, optional): Describes which channels to include or discard.
                Defaults to None; all channels will be used.
            poststimulus_length (float, optional): [description]. Defaults to 0.5.
            target_label (str): label of target symbol. Defaults to "target"

        Returns
        -------
            trial_data (np.ndarray): shape (channels, trials, samples) reshaped data
            labels (np.ndarray): integer label for each trial
        """
        # Remove the channels that we are not interested in
        if channel_map:
            channels_to_remove = [idx for idx, value in enumerate(channel_map) if value == 0]
            eeg_data = np.delete(eeg_data, channels_to_remove, axis=0)

        # Number of samples we are interested per trial
        poststim_samples = int(poststimulus_length * sample_rate)
        prestim_samples = int(prestimulus_length * sample_rate)

        # triggers in seconds are mapped to triggers in number of samples.
        triggers = list(map(lambda x: int((x + offset) * sample_rate), timing_info))

        # Label for every trial in 0 or 1
        targetness_labels = np.zeros(len(triggers), dtype=np.compat.long)
        reshaped_trials = []
        for trial_idx, (trial_label, trigger) in enumerate(zip(trial_targetness_label, triggers)):
            if trial_label == target_label:
                targetness_labels[trial_idx] = 1

            # For every channel append filtered channel data to trials
            reshaped_trials.append(eeg_data[:, trigger - prestim_samples: trigger + poststim_samples])

        return np.stack(reshaped_trials, 1), targetness_labels


def update_inquiry_timing(timing: List[List[float]], downsample: int) -> List[List[float]]:
    """Update inquiry timing to reflect downsampling."""

    for i, inquiry in enumerate(timing):
        for j, time in enumerate(inquiry):
            timing[i][j] = time // downsample

    return timing


def mne_epochs(mne_data: RawArray, trigger_timing: List[float],
               trial_length: float, trigger_labels: List[int]) -> Epochs:
    """MNE Epochs.

    Using an MNE RawArray, reshape the data given trigger information. If two labels present [0, 1],
    each may be accessed by numbered order. Ex. first_class = epochs['1'], second_class = epochs['2']
    """
    annotations = Annotations(trigger_timing, [trial_length] * len(trigger_timing), trigger_labels)
    mne_data.set_annotations(annotations)

    events_from_annot, _ = mne.events_from_annotations(mne_data)
    return Epochs(mne_data, events_from_annot)


def alphabetize(stimuli: List[str]) -> List[str]:
    """Alphabetize.

    Given a list of string stimuli, return a list of sorted stimuli by alphabet.
    """
    return sorted(stimuli, key=lambda x: re.sub(r'[^a-zA-Z0-9 \n\.]', 'ZZ', x).lower())


def inq_generator(query: List[str],
                  timing: List[float] = [1, 0.2],
                  color: List[str] = ['red', 'white'],
                  inquiry_count: int = 1,
                  stim_jitter: float = 0,
                  stim_order: StimuliOrder = StimuliOrder.RANDOM,
                  is_txt: bool = True) -> InquirySchedule:
    """Given the query set, prepares the stimuli, color and timing

    Parameters
    ----------
        query(list[str]): list of queries to be shown
        timing(list[float]): Task specific timing for generator
        color(list[str]): Task specific color for generator
            First element is the target, second element is the fixation
            Observe that [-1] element represents the trial information
    Return
    ------
        schedule_inq(tuple(
            samples[list[list[str]]]: list of inquiries
            timing(list[list[float]]): list of timings
            color(list(list[str])): list of colors)): scheduled inquiries
    """

    if stim_order == StimuliOrder.ALPHABETICAL:
        query = alphabetize(query)
    else:
        random.shuffle(query)

    stim_length = len(query)

    # Init some lists to construct our stimuli with
    samples, times, colors = [], [], []
    for _ in range(inquiry_count):

        # append a fixation cross. if not text, append path to image fixation
        sample = [get_fixation(is_txt)]

        # construct the sample from the query
        sample += [i for i in query]
        samples.append(sample)

        times.append([timing[i] for i in range(len(timing) - 1)])
        base_timing = timing[-1]
        times[-1] += jittered_timing(base_timing, stim_jitter, stim_length)

        # append colors
        colors.append([color[i] for i in range(len(color) - 1)] +
                      [color[-1]] * stim_length)
    return InquirySchedule(samples, times, colors)


def best_selection(selection_elements: list,
                   val: list,
                   len_query: int,
                   always_included: List[str] = None) -> list:
    """Best Selection.

    Given set of elements and a value function over the set, picks the len_query
        number of elements with the best value.

    Parameters
    ----------
        selection_elements(list[str]): the set of elements
        val(list[float]): values for the corresponding elements
        len_query(int): number of elements to be picked from the set
        always_included(list[str]): subset of elements that should always be
            included in the result. Defaults to None.
    Return
    ------
        best_selection(list[str]): elements from selection_elements with the best values
    """

    always_included = always_included or []
    # pick the top n items sorted by value in decreasing order
    elem_val = dict(zip(selection_elements, val))
    best = sorted(selection_elements, key=elem_val.get, reverse=True)[0:len_query]

    replacements = [
        item for item in always_included
        if item not in best and item in selection_elements
    ][0:len_query]

    if replacements:
        best[-len(replacements):] = replacements
    return best


def best_case_rsvp_inq_gen(alp: list,
                           session_stimuli: np.ndarray,
                           timing: List[float] = [1, 0.2],
                           color: List[str] = ['red', 'white'],
                           stim_number: int = 1,
                           stim_length: int = 10,
                           stim_order: StimuliOrder = StimuliOrder.RANDOM,
                           is_txt: bool = True,
                           inq_constants: List[str] = None) -> InquirySchedule:
    """Best Case RSVP Inquiry Generation.

    Generates RSVPKeyboard inquiry by picking n-most likely letters.

    Parameters
    ----------
        alp(list[str]): alphabet (can be arbitrary)
        session_stimuli(ndarray[float]): quantifier metric for query selection
            dim(session_stimuli) = card(alp)!
        timing(list[float]): Task specific timing for generator
        color(list[str]): Task specific color for generator
            First element is the target, second element is the fixation
            Observe that [-1] element represents the trial information
        stim_number(int): number of random stimuli to be created
        stim_length(int): number of trials in a inquiry
        stim_order(StimuliOrder): ordering of stimuli in the inquiry
        inq_constants(list[str]): list of letters that should always be
            included in every inquiry. If provided, must be alp items.
    Return
    ------
        schedule_inq(tuple(
            samples[list[list[str]]]: list of inquiries
            timing(list[list[float]]): list of timings
            color(list(list[str])): list of colors)): scheduled inquiries
    """

    if len(alp) != len(session_stimuli):
        raise BciPyCoreException((
            f'Missing information about alphabet.'
            f'len(alp):{len(alp)} and len(session_stimuli):{len(session_stimuli)} should be same!'))

    if inq_constants and not set(inq_constants).issubset(alp):
        raise BciPyCoreException('Inquiry constants must be alphabet items.')

    # query for the best selection
    query = best_selection(
        alp,
        session_stimuli,
        stim_length,
        inq_constants)

    if stim_order == StimuliOrder.ALPHABETICAL:
        query = alphabetize(query)
    else:
        random.shuffle(query)

    # Init some lists to construct our stimuli with
    samples, times, colors = [], [], []
    for _ in range(stim_number):

        # append a fixation cross. if not text, append path to image fixation
        sample = [get_fixation(is_txt)]

        # construct the sample from the query
        sample += [i for i in query]
        samples.append(sample)

        # append timing
        times.append([timing[i] for i in range(len(timing) - 1)] +
                     [timing[-1]] * stim_length)

        # append colors
        colors.append([color[i] for i in range(len(color) - 1)] +
                      [color[-1]] * stim_length)
    return InquirySchedule(samples, times, colors)


def calibration_inquiry_generator(
        alp: List[str],
        timing: List[float] = [0.5, 1, 0.2],
        jitter: Optional[int] = None,
        color: List[str] = ['green', 'red', 'white'],
        stim_number: int = 10,
        stim_length: int = 10,
        stim_order: StimuliOrder = StimuliOrder.RANDOM,
        target_positions: TargetPositions = TargetPositions.RANDOM,
        nontarget_inquiries: int = 10,
        is_txt: bool = True) -> InquirySchedule:
    """Calibration Inquiry Generator.

    Generates inquiries with target letters in all possible positions.

    Parameters
    ----------
        alp(list[str]): stimuli
        timing(list[float]): Task specific timing for generator.
            [target, fixation, stimuli]
        jitter(int): jitter for stimuli timing. If None, no jitter is applied.
        color(list[str]): Task specific color for generator
            [target, fixation, stimuli]
        stim_number(int): number of trials in a inquiry
        stim_length(int): number of random stimuli to be created
        stim_order(StimuliOrder): ordering of stimuli in the inquiry
        target_positions(TargetPositions): positioning of targets to select for the inquiries
        nontarget_inquiries(int): percentage of inquiries for which target letter flashed is not in inquiry
        is_txt(bool): whether or not the stimuli type is text. False would be an image stimuli.

    Return
    ------
        schedule_inq(tuple(
            samples[list[list[str]]]: list of inquiries
            timing(list[list[float]]): list of timings
            color(list(list[str])): list of colors)): scheduled inquiries
    """
    assert len(timing) == 3, "timing must include values for [target, fixation, stimuli]"
    target_indexes = []
    no_target = None

    if (target_positions == target_positions.DISTRIBUTED):
        target_indexes = distributed_target_positions(stim_number, stim_length, nontarget_inquiries)
    else:
        # make list of random targets with correct number of non-target inquiries
        num_nontarget_inquiry = int((nontarget_inquiries / 100) * stim_number)
        num_target_inquiry = stim_number - num_nontarget_inquiry
        target_indexes = [no_target] * num_nontarget_inquiry
        target_indexes.extend(random.choices(range(stim_length), k=num_target_inquiry))
        random.shuffle(target_indexes)

    samples, times, colors = [], [], []

    for i in range(stim_number):
        inquiry = random.sample(alp, k=stim_length)

        if stim_order == StimuliOrder.ALPHABETICAL:
            inquiry = alphabetize(inquiry)

        target_index = target_indexes[i]
        if target_index is no_target:
            target = random.choice(list(set(alp) - set(inquiry)))
        else:
            target = inquiry[target_index]

        sample = [target, get_fixation(is_txt=is_txt), *inquiry]

        samples.append(sample)

        # timing for fixation and prompt
        init_timing = [timing[i] for i in range(len(timing) - 1)]
        # pull out timing for the inquiry stimuli
        stim_time = timing[-1]
        if jitter:
            _inq_timing = jittered_timing(stim_time, jitter, stim_length)
            inq_timing = init_timing + _inq_timing
        else:
            inq_timing = init_timing + [stim_time] * stim_length
        times.append(inq_timing)
        colors.append([color[i] for i in range(len(color) - 1)] +
                      [color[-1]] * stim_length)

    return InquirySchedule(samples, times, colors)


def jittered_timing(time: float, jitter: float, stim_count: int) -> List[float]:
    """Jittered timing.

    Using a base time and a jitter, generate a list (with length stim_count) of timing that is uniformly distributed.
    """
    assert time > jitter, (
        f'Jitter timing [{jitter}] must be less than stimuli timing =[{time}] in the inquiry.')
    return np.random.uniform(low=time - jitter, high=time + jitter, size=(stim_count,)).tolist()


def distributed_target_positions(stim_number: int, stim_length: int, nontarget_inquiries: int) -> list:
    """Distributed Target Positions.

    Generates evenly distributed target positions, including target letter not flashed at all, and shuffles them.
    Args:
        stim_number(int): Number of trials in calibration
        stim_length(int): Number of stimuli in each inquiry
        nontarget_inquiries(int): percentage of inquiries for which target letter flashed is not in inquiry

    Return distributed_target_positions(list): targets: array of target indexes to be chosen
    """

    targets = []
    no_target = None

    # find number of target and nontarget inquiries
    num_nontarget_inquiry = int(stim_number * (nontarget_inquiries / 100))
    num_target_inquiry = stim_number - num_nontarget_inquiry

    # find number each target position is repeated, and remaining number
    num_pos = (int)(num_target_inquiry / stim_length)
    num_rem_pos = (num_target_inquiry % stim_length)

    # add correct number of None's for nontarget inquiries
    targets = [no_target] * num_nontarget_inquiry

    # add distributed list of target positions
    targets.extend(list(range(stim_length)) * num_pos)

    # pick leftover positions randomly
    rem_pos = list(range(stim_length))
    random.shuffle(rem_pos)
    rem_pos = rem_pos[0:num_rem_pos]
    targets.extend(rem_pos)

    # shuffle targets
    random.shuffle(targets)

    return targets


def get_task_info(experiment_length: int, task_color: str) -> Tuple[List[str], List[str]]:
    """Get Task Info.

    Generates fixed RSVPKeyboard task text and color information for
            display.
    Args:
        experiment_length(int): Number of inquiries for the experiment
        task_color(str): Task information display color

    Return get_task_info((tuple): task_text: array of task text to display
                   task_color: array of colors for the task text
                   )
    """

    # Do list comprehensions to get the arrays for the task we need.
    task_text = ['%s/%s' % (stim + 1, experiment_length)
                 for stim in range(experiment_length)]
    task_color = [[str(task_color)] for stim in range(experiment_length)]

    return (task_text, task_color)


def resize_image(image_path: str, screen_size: tuple, sti_height: int) -> Tuple[float, float]:
    """Resize Image.

    Returns the width and height that a given image should be displayed at
    given the screen size, size of the original image, and stimuli height
    parameter"""
    # Retrieve image width and height
    with Image.open(image_path) as pillow_image:
        image_width, image_height = pillow_image.size

    # Resize image so that its largest dimension is the stimuli size defined
    # in the parameters file
    if image_width >= image_height:
        proportions = (1, (image_height / image_width))
    else:
        proportions = ((image_width / image_height), 1)

    # Adjust image size to scale with monitor size
    screen_width, screen_height = screen_size
    if screen_width >= screen_height:
        sti_size = ((screen_height / screen_width) * sti_height *
                    proportions[0], sti_height * proportions[1])
    else:
        sti_size = (
            sti_height * proportions[0],
            (screen_width / screen_height) * sti_height * proportions[1])

    return sti_size


def play_sound(sound_file_path: str,
               dtype: str = 'float32',
               track_timing: bool = False,
               sound_callback=None,
               sound_load_buffer_time: float = 0.5,
               experiment_clock=None,
               trigger_name: str = None,
               timing: list = []) -> list:
    """Play Sound.

    Using soundevice and soundfile, play a sound giving options to buffer times between
        loading sound into memory and after playing. If desired, marker writers or list based
        timing with psychopy clocks may be passed and sound timing returned.


    PARAMETERS
    ----------
    :param: sound_file_path
    :param: dtype: type of sound ex. float32.
    :param: track_timing: whether or not to track timing of sound playin
    :param: sound_callback: trigger based callback (see MarkerWriter and NullMarkerWriter)
    :param: sound_load_buffer_time: time to wait after loading file before playing
    :param: experiment_clock: psychopy clock to get time of sound stimuli
    :param: trigger_name: name of the sound trigger
    :param: timing: list of triggers in the form of trigger name, trigger timing
    :resp: timing
    """

    try:
        # load in the sound file and wait some time before playing
        data, fs = sf.read(sound_file_path, dtype=dtype)
        core.wait(sound_load_buffer_time)

    except Exception as e:
        error_message = f'Sound file could not be found or initialized. \n Exception={e}'
        log.exception(error_message)
        raise BciPyCoreException(error_message)

    #  if timing is wanted, get trigger timing for this sound stimuli
    if track_timing:
        # if there is a timing callback for sound, evoke it
        if sound_callback is not None:
            sound_callback(experiment_clock, trigger_name)
        timing.append([trigger_name, experiment_clock.getTime()])

    # play our loaded sound and wait for some time before it's finished
    # NOTE: there is a measurable delay for calling sd.play. (~ 0.1 seconds;
    # which I believe happens prior to the sound playing).
    sd.play(data, fs)
    # sd.play returns immediately (according to the docs); wait for the duration of the sound
    duration = len(data) / fs
    core.wait(duration)
    return timing


def soundfiles(directory: str) -> Iterator[str]:
    """Creates a generator that cycles through sound files (.wav) in a
    directory and returns the path to next sound file on each iteration.

    Parameters:
    -----------
        directory - path to the directory which contains .wav files
    Returns:
    --------
        iterator that infinitely cycles through the filenames.
    """
    if not path.isdir(directory):
        error_message = f'Invalid directory=[{directory}] for sound files.'
        log.error(error_message)
        raise BciPyCoreException(error_message)
    if not directory.endswith(sep):
        directory += sep
    return itertools.cycle(glob.glob(directory + '*.wav'))


def get_fixation(is_txt: bool) -> str:
    """Get Fixation.

    Return the correct stimulus fixation given the type (text or image).
    """
    if is_txt:
        return DEFAULT_TEXT_FIXATION
    else:
        return DEFAULT_FIXATION_PATH


def ssvep_to_code(refresh_rate: int = 60, flicker_rate: int = 10) -> List[int]:
    """Convert SSVEP to Code.

    Converts a SSVEP (steady state visual evoked potential; ex. 10 Hz) to a code (0,1)
    given the refresh rate of the monitor (Hz) provided and a desired flicker rate (Hz).

    Parameters:
    -----------
        refresh_rate: int, refresh rate of the monitor (Hz)
        flicker_rate: int, desired flicker rate (Hz)
    Returns:
    --------
        list of 0s and 1s that represent the code for the SSVEP on the monitor.
    """
    if flicker_rate > refresh_rate:
        raise BciPyCoreException('flicker rate cannot be greater than refresh rate')
    if flicker_rate <= 1:
        raise BciPyCoreException('flicker rate must be greater than 1')

    # get the number of frames per flicker
    length_flicker = refresh_rate / flicker_rate

    if length_flicker.is_integer():
        length_flicker = int(length_flicker)
    else:
        err_message = f'flicker rate={flicker_rate} is not an integer multiple of refresh rate={refresh_rate}'
        log.exception(err_message)
        raise BciPyCoreException(err_message)

    # start the first frames as off (0) for length of flicker;
    # it will then toggle on (1)/ off (0) for length of flicker until all frames are filled for refresh rate.
    t = 0
    codes = []
    for _ in range(flicker_rate):
        codes += [t] * length_flicker
        t = 1 - t

    return codes



from queue import Queue, Empty
import itertools as it


class DataSource:
    """Abstract base class for an object with data. Provided to the EEGFrame
    for streaming data."""

    def next(self):
        """Provide the next record."""
        raise NotImplementedError('Subclass must define the next method')

    def next_n(self, n: int, fast_forward=False):
        """Provides the next n records as a list"""
        raise NotImplementedError('Subclass must define the next_n method')


class QueueDataSource(DataSource):
    """Queue backed data source."""

    def __init__(self, q: Queue):
        self.queue = q
        self.wait = 0.1

    def next(self):
        """Provide the next record."""
        try:
            return self.queue.get()
        except Empty:
            raise StopIteration

    def next_n(self, n: int, fast_forward=False):
        """Provides the next n records as a list"""
        data = []
        while len(data) < n:
            try:
                record = self.queue.get(True, self.wait)
                data.append(record)
            except Empty:
                raise StopIteration
        return data


class GeneratorDataSource(DataSource):
    """DataSource that uses a provided python generator."""

    def __init__(self, gen):
        self.gen = gen

    def next(self):
        """Provide the next record."""
        return next(self.gen)

    def next_n(self, n, fast_forward=False):
        """Provides the next n records as a list"""
        return list(it.islice(self.gen, n))



"""Tests for Parameter related functionality"""
import shutil
import tempfile
import unittest
from collections import abc
from pathlib import Path

from bcipy.config import DEFAULT_PARAMETERS_PATH
from bcipy.helpers.parameters import Parameters


class TestParameters(unittest.TestCase):
    """Tests for loading and saving Parameters."""

    def setUp(self):
        """Override; set up the needed path for load functions."""

        self.parameters_location = DEFAULT_PARAMETERS_PATH
        self.temp_dir = tempfile.mkdtemp()

    def tearDown(self):
        """Override"""
        shutil.rmtree(self.temp_dir)

    def test_load_json_parameters_returns_dict(self):
        """Test load parameters returns a Python dict."""

        parameters = Parameters(source=self.parameters_location)
        self.assertTrue(isinstance(parameters, abc.MutableMapping))

    def test_load_json_parameters_throws_error_on_wrong_path(self):
        """Test load parameters returns error on entering wrong path."""

        with self.assertRaises(Exception):
            Parameters(source='/garbage/dir/wont/work')

    def test_load_initializes_parameters(self):
        """Test load parameters initializes data."""
        parameters = Parameters(source=self.parameters_location)
        self.assertTrue(len(parameters) > 0)

    def test_initialization_without_source(self):
        """Test parameter initialization without a JSON source file."""

        parameters = Parameters(source=None)
        self.assertTrue(isinstance(parameters, abc.MutableMapping))
        self.assertEqual(len(parameters), 0)

    def test_load_data(self):
        """Test load data."""
        sample_config = {
            "fake_data": {
                "value": "true",
                "section": "bci_config",
                "readableName": "Fake EEG Data On/Off",
                "helpTip":
                "If ‘true’, fake EEG data will be used instead of real EEG data.",
                "recommended_values": "",
                "type": "bool"
            },
            "acq_device": {
                "value": "LSL",
                "section": "acq_config",
                "readableName": "Acquisition Device Connection Method",
                "helpTip":
                "Specifies the method used to connect to the data acquisition device (LSL or DSI).",
                "recommended_values": ["DSI", "LSL"],
                "type": "str"
            }
        }
        parameters = Parameters(source=None)
        parameters.load(sample_config)
        self.assertEqual(len(parameters), 2)

    def test_load_data_with_missing_field(self):
        """Test that an exception is raised if data has a missing field."""
        data_with_missing_help_tip = {
            "fake_data": {
                "value": "true",
                "section": "bci_config",
                "readableName": "Fake EEG Data On/Off",
                "recommended_values": "",
                "type": "bool"
            }
        }
        parameters = Parameters(source=None)
        with self.assertRaises(Exception):
            parameters.load(data_with_missing_help_tip)

    def test_load_data_with_unsupported_type(self):
        """Test that an exception is raised if data has an unsupported data type."""
        data_with_unsupported_type = {
            "fake_data": {
                "value": "true",
                "section": "bci_config",
                "readableName": "Fake EEG Data On/Off",
                "helpTip": "",
                "recommended_values": "",
                "type": "custom_type"
            }
        }
        parameters = Parameters(source=None)
        with self.assertRaises(Exception):
            parameters.load(data_with_unsupported_type)

    def test_cast_values(self):
        """Test cast_values."""

        sample_config = {
            "myint": {
                "value": "1",
                "section": "",
                "readableName": "",
                "helpTip": "",
                "recommended_values": "",
                "type": "int"
            },
            "mybool": {
                "value": "true",
                "section": "",
                "readableName": "",
                "helpTip": "",
                "recommended_values": "",
                "type": "bool"
            },
            "mypath": {
                "value": "bcipy/parameters/parameters.json",
                "section": "",
                "readableName": "",
                "helpTip": "",
                "recommended_values": "",
                "type": "directorypath"
            },
            "mystr": {
                "value": "hello",
                "section": "",
                "readableName": "",
                "helpTip": "",
                "recommended_values": "",
                "type": "str"
            },
        }
        parameters = Parameters(source=None, cast_values=True)
        parameters.load(sample_config)

        self.assertEqual(parameters['myint'], 1)
        self.assertEqual(parameters.get('myint', 0), 1,
                         'Should support `get` method with a default value')
        self.assertEqual(
            parameters.get('myint'), 1,
            'Should support `get` method without a default value')

        self.assertEqual(parameters['mybool'], True)
        self.assertEqual(parameters.get('mybool', False), True)
        self.assertEqual(parameters['mypath'],
                         'bcipy/parameters/parameters.json')
        self.assertEqual(parameters['mystr'], 'hello')
        self.assertEqual(parameters.get('mystr'), 'hello')
        self.assertFalse(parameters.get('missing_param', False))

        parameters.cast_values = False
        self.assertEqual(parameters['myint']['value'], '1')
        self.assertEqual(parameters['mybool']['value'], 'true')

    def test_setting_valid_values(self):
        """Test that new parameters can be added"""
        parameters = Parameters(source=None, cast_values=False)
        self.assertEqual(len(parameters), 0)
        parameters['mystr'] = {
            "value": "hello",
            "section": "",
            "readableName": "",
            "helpTip": "",
            "recommended_values": "",
            "type": "str"
        }
        self.assertEqual(len(parameters), 1)

    def test_setting_invalid_values(self):
        """Test that directly setting invalid values raises an exception."""
        missing_help_tip = {
            "value": "true",
            "section": "",
            "readableName": "",
            "recommended_values": "",
            "type": "bool"
        }
        unsupported_type = {
            "value": "true",
            "section": "bci_config",
            "readableName": "Fake EEG Data On/Off",
            "helpTip": "",
            "recommended_values": "",
            "type": "custom_type"
        }

        parameters = Parameters(source=None, cast_values=False)
        self.assertEqual(len(parameters), 0)

        with self.assertRaises(Exception):
            parameters['test1'] = missing_help_tip

        with self.assertRaises(Exception):
            parameters['test2'] = unsupported_type

    def test_updating_uncast_values(self):
        """Test that new parameters can be added"""
        parameters = Parameters(source=None, cast_values=False)
        parameters['mystr'] = {
            "value": "hello",
            "section": "",
            "readableName": "",
            "helpTip": "",
            "recommended_values": "",
            "type": "str"
        }
        parameters['mystr']['value'] = 'hello world'
        self.assertEqual(parameters['mystr']['value'], 'hello world')

    def test_setting_existing_cast_values(self):
        """Test setting data when Parameters are cast"""
        sample_config = {
            "acq_port": {
                "value": "8000",
                "section": "acquisition",
                "readableName": "Acquisition Port",
                "helpTip": "",
                "recommended_values": "",
                "type": "int"
            },
            "acq_device": {
                "value": "LSL",
                "section": "acquisition",
                "readableName": "Acquisition Device",
                "helpTip": "",
                "recommended_values": ["LSL", "DSI"],
                "type": "str"
            },
            "is_txt_stim": {
                "value": "false",
                "section": "",
                "readableName": "",
                "helpTip": "",
                "recommended_values": "",
                "type": "bool"
            }
        }
        parameters = Parameters(source=None, cast_values=True)
        parameters.load(sample_config)
        self.assertEqual(parameters['is_txt_stim'], False)

        parameters['acq_port'] = 9000
        parameters['acq_device'] = 'DSI'
        parameters['is_txt_stim'] = True
        print(parameters)
        self.assertEqual(parameters['acq_port'], 9000)
        self.assertEqual(parameters['acq_device'], 'DSI')
        self.assertEqual(parameters['is_txt_stim'], True)

        parameters.cast_values = False
        self.assertEqual(parameters['acq_port']['value'], '9000')
        self.assertEqual(parameters['acq_device']['value'], 'DSI')
        self.assertEqual(parameters['is_txt_stim']['value'], 'true')

    def test_update(self):
        """Test update method"""
        sample_config = {
            "acq_port": {
                "value": "8000",
                "section": "acquisition",
                "readableName": "Acquisition Port",
                "helpTip": "",
                "recommended_values": "",
                "type": "int"
            },
            "acq_device": {
                "value": "LSL",
                "section": "acquisition",
                "readableName": "Acquisition Device",
                "helpTip": "",
                "recommended_values": ["LSL", "DSI"],
                "type": "str"
            }
        }
        parameters = Parameters(source=None, cast_values=True)
        parameters.load(sample_config)
        parameters.update(acq_port=9000, acq_device='DSI')

        self.assertEqual(parameters['acq_port'], 9000)
        self.assertEqual(parameters['acq_device'], 'DSI')

    def test_setting_missing_cast_values(self):
        """Test setting cast data for values that have not already been loaded."""
        parameters = Parameters(source=None, cast_values=True)
        with self.assertRaises(Exception):
            parameters['acq_port'] = 9000

    def test_values(self):
        """Test values method"""
        sample_config = {
            "acq_port": {
                "value": "8000",
                "section": "acquisition",
                "readableName": "Acquisition Port",
                "helpTip": "",
                "recommended_values": "",
                "type": "int"
            },
            "acq_device": {
                "value": "LSL",
                "section": "acquisition",
                "readableName": "Acquisition Device",
                "helpTip": "",
                "recommended_values": ["LSL", "DSI"],
                "type": "str"
            }
        }
        parameters = Parameters(source=None, cast_values=True)
        parameters.load(sample_config)
        self.assertEqual(list(parameters.keys()), ['acq_port', 'acq_device'])
        self.assertEqual(list(parameters.values()), [8000, 'LSL'])

        parameters.cast_values = False
        for value in parameters.values():
            self.assertEqual(type(value), dict)

    def test_save_as(self):
        """Test saving data to a new json file"""

        parameters = Parameters(source=self.parameters_location)
        name = 'parameters_copy.json'
        parameters.save(directory=self.temp_dir, name=name)

        source = str(Path(self.temp_dir, name))
        params2 = Parameters(source=source, cast_values=False)

        self.assertEqual(parameters, params2)

    def test_save(self):
        """Test saving to overwrite source file."""

        # Save a copy to the temp directory
        parameters = Parameters(source=self.parameters_location,
                                cast_values=True)
        name = 'parameters_copy.json'
        parameters.save(directory=self.temp_dir, name=name)

        source = str(Path(self.temp_dir, name))
        params2 = Parameters(source=source, cast_values=True)
        self.assertEqual(parameters, params2)

        # Change a parameter and overwrite
        self.assertTrue(parameters['parameter_location'] != source)
        params2['parameter_location'] = source
        params2.save()

        # Load from the save file and confirm that saved values persisted
        params3 = Parameters(source=source, cast_values=True)
        self.assertEqual(params3['parameter_location'], source)
        self.assertEqual(params3, params2)
        self.assertNotEqual(params3, parameters)

    def test_save_new(self):
        """Test saving a new params file."""
        parameters = Parameters(source=None, cast_values=False)
        parameters['mystr'] = {
            "value": "hello",
            "section": "",
            "readableName": "",
            "helpTip": "",
            "recommended_values": "",
            "type": "str"
        }
        with self.assertRaises(Exception):
            # Missing directory and file name
            parameters.save()

        with self.assertRaises(Exception):
            parameters.save(directory=self.temp_dir)
        with self.assertRaises(Exception):
            parameters.save(name='my_params.json')

        parameters.save(directory=self.temp_dir, name='my_params.json')

    def test_items(self):
        """Test items"""
        parameters = Parameters(source=None, cast_values=False)
        parameters['mystr'] = {
            "value": "hello",
            "section": "",
            "readableName": "",
            "helpTip": "",
            "recommended_values": "",
            "type": "str"
        }
        self.assertEqual(len(parameters.items()), 1)
        for key, val in parameters.items():
            self.assertEqual(key, 'mystr')
            self.assertEqual(type(val), dict)

        parameters.cast_values = True
        self.assertEqual(len(parameters.items()), 1)
        for key, val in parameters.items():
            self.assertEqual(key, 'mystr')
            self.assertEqual(val, 'hello')

    def test_copy(self):
        """Test copy"""
        parameters = Parameters(source=self.parameters_location,
                                cast_values=True)
        params_copy = parameters.copy()

        self.assertEqual(params_copy.source, None)
        self.assertEqual(params_copy.cast_values, parameters.cast_values)
        self.assertEqual(params_copy.values(), parameters.values())

    def test_check_entry(self):
        parameters = Parameters(source=None, cast_values=False)
        parameters.check_valid_entry(
            "fake_data", {
                "value": "true",
                "section": "bci_config",
                "readableName": "Fake Data Sessions",
                "helpTip": "If true, fake data server used",
                "recommended_values": "",
                "type": "bool"
            })
        with self.assertRaises(Exception):
            parameters.check_valid_entry("fake_data", True)

    def test_check_entry_bool_type(self):
        """Test that invalid bool types are rejected"""
        parameters = Parameters(source=None, cast_values=False)
        with self.assertRaises(Exception):
            parameters.check_valid_entry(
                "fake_data", {
                    "value": True,
                    "section": "bci_config",
                    "readableName": "Fake Data Sessions",
                    "helpTip": "If true, fake data server used",
                    "recommended_values": "",
                    "type": "bool"
                })

    def test_alternate_constructor(self):
        """Test alternate constructor from cast values"""
        parameters = Parameters.from_cast_values(myint=1,
                                                 mybool=True,
                                                 mystr="Testing")
        self.assertTrue(parameters.cast_values)
        self.assertEqual(parameters['myint'], 1)
        self.assertEqual(parameters['mybool'], True)
        self.assertEqual(parameters['mystr'], 'Testing')

        parameters.cast_values = False
        self.assertEqual(parameters['myint']['value'], '1')
        self.assertEqual(parameters['mybool']['value'], 'true')
        self.assertEqual(parameters['mystr']['value'], 'Testing')

    def test_add_missing(self):
        """Test add_missing_items"""
        entry1 = {
            "value": "8000",
            "section": "acquisition",
            "readableName": "Acquisition Port",
            "helpTip": "",
            "recommended_values": "",
            "type": "int"
        }
        entry2 = {
            "value": "LSL",
            "section": "acquisition",
            "readableName": "Acquisition Device",
            "helpTip": "",
            "recommended_values": ["LSL", "DSI"],
            "type": "str"
        }

        parameters = Parameters(source=None)
        parameters.load({"acq_port": entry1})

        new_params = Parameters(source=None)
        new_params.load({"acq_port": entry1, "acq_device": entry2})

        self.assertFalse('acq_device' in parameters.keys())
        self.assertTrue(parameters.add_missing_items(new_params))
        self.assertTrue('acq_device' in parameters.keys())

        self.assertFalse(parameters.add_missing_items(new_params),
                         "Only new parameters should be added.")

    def test_changed_parameters(self):
        """Test diff calculations"""
        entry1 = {
            "value": "8000",
            "section": "acquisition",
            "readableName": "Acquisition Port",
            "helpTip": "",
            "recommended_values": "",
            "type": "int"
        }
        entry2 = {
            "value": "75E+6",
            "section": "artifact_rejection",
            "readableName": "High Voltage Threshold Value",
            "helpTip": "Specifies the high voltage threshold (in microvolts)",
            "recommended_values": "",
            "type": "float"
        }
        entry2_same = {
            "value": "75000000.0",
            "section": "artifact_rejection",
            "readableName": "High Voltage Threshold Value",
            "helpTip": "Specifies the high voltage threshold (in microvolts)",
            "recommended_values": "",
            "type": "float"
        }
        entry3 = {
            "value": "DSI-24",
            "section": "acquisition",
            "readableName": "Acquisition Device",
            "helpTip": "",
            "recommended_values": ["DSI-24", "DSI-VR300"],
            "type": "str"
        }
        entry3_modified = {
            "value": "DSI-VR300",
            "section": "acquisition",
            "readableName": "Acquisition Device",
            "helpTip": "",
            "recommended_values": ["DSI-24", "DSI-VR300"],
            "type": "str"
        }
        parameters = Parameters(source=None)
        parameters.load({
            "entry_1": entry1,
            "entry_2": entry2,
            "entry_3": entry3
        })

        new_params = Parameters(source=None)
        new_params.load({
            "entry_1": entry1,
            "entry_2": entry2_same,
            "entry_3": entry3_modified
        })
        changes = new_params.diff(parameters)

        self.assertEqual(len(changes.keys()), 1)
        self.assertTrue("entry_3" in changes.keys())
        self.assertEqual(changes["entry_3"].original_value, entry3['value'])
        self.assertEqual(changes["entry_3"].parameter['value'],
                         entry3_modified['value'])


if __name__ == '__main__':
    unittest.main()



from bcipy.signal.model.base_model import SignalModel, ModelEvaluationReport
from bcipy.signal.model.pca_rda_kde.pca_rda_kde import PcaRdaKdeModel
from bcipy.signal.model.rda_kde.rda_kde import RdaKdeModel

__all__ = [
    "SignalModel",
    "PcaRdaKdeModel",
    "RdaKdeModel",
    "ModelEvaluationReport",
]




class BciPyCoreException(Exception):
    """BciPy Core Exception.

    Thrown when an error occurs specific to BciPy core concepts.
    """

    def __init__(self, message, errors=None):
        super().__init__(message)
        self.message = message
        self.errors = errors


class SignalException(BciPyCoreException):
    """
    Signal Exception.

    Thrown when signal module encounters an error.
    """

    def __init__(self, message, errors=None):
        super().__init__(message)
        self.errors = errors


class FieldException(BciPyCoreException):
    """Field Exception.

    Thrown when there is an exception relating to experimental fields.
    """
    ...


class ExperimentException(BciPyCoreException):
    """Experiment Exception.

    Thrown when there is an exception relating to experiments.
    """
    ...


class UnregisteredExperimentException(ExperimentException):
    """Unregistered Experiment.

    Thrown when experiment is not registered in the provided experiment path.
    """

    ...


class UnregisteredFieldException(FieldException):
    """Unregistered Field.

    Thrown when field is not registered in the provided field path.
    """

    ...


class InvalidExperimentException(ExperimentException):
    """Invalid Experiment Exception.

    Thrown when providing experiment data in the incorrect format.
    """

    ...


class InvalidFieldException(FieldException):
    """Invalid Field Exception.

    Thrown when providing field data in the incorrect format.
    """

    ...


class UnsupportedResponseType(BciPyCoreException):
    """Unsupported ResponseType

    Thrown when attempting to set the response type of a language model to an
    unsupported value."""
    ...


class TaskConfigurationException(BciPyCoreException):
    """Task Configuration Exception.

    Thrown when attempting to run a task with invalid configurations"""
    ...


class InvalidLanguageModelException(BciPyCoreException):
    """Invalid Language Model Exception.

    Thrown when attempting to load a language model from an invalid path"""
    ...


class KenLMInstallationException(BciPyCoreException):
    """KenLM Installation Exception.

    Thrown when attempting to import kenlm without installing the module"""
    ...



import os
import sys
import platform
from shutil import rmtree

from setuptools import find_packages, setup, Command

# Package meta-data.
NAME = 'bcipy'
DESCRIPTION = 'Python Software for Brain-Computer Interface.'
URL = 'https://github.com/CAMBI-tech/BciPy'
EMAIL = 'cambi_support@googlegroups.com'
AUTHOR = 'CAMBI'
REQUIRES_PYTHON = '>3.7,<3.10'

VERSION = '2.0.0rc3'

REQUIRED = []

# What packages are required for this module to be executed?
if platform.system() == 'Windows' or platform.system() == 'Darwin':
    with open('requirements-winmac.txt', 'r', encoding='utf-8') as f:
        REQUIRED += f.read().splitlines()
with open('requirements.txt', 'r', encoding='utf-8') as f:
    REQUIRED += f.read().splitlines()

# The rest you shouldn't have to touch too much :)
# ------------------------------------------------
# Except, perhaps the License and Trove Classifiers!
# If you do change the License, remember to change the Trove Classifier for that!

here = os.path.abspath(os.path.dirname(__file__))

long_description = 'Python Brain-Computer Interface Software'

about = {'__version__': VERSION}


class UploadCommand(Command):
    """Support setup.py upload.
    
    Modified from https://github.com/kennethreitz/setup.py
    """

    description = 'Build and publish the package.'
    user_options = []

    @staticmethod
    def status(s):
        """Prints things in bold."""
        print('\033[1m{0}\033[0m'.format(s))

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        try:
            self.status('Removing previous builds…')
            rmtree(os.path.join(here, 'dist'))
        except OSError:
            pass

        self.status('Building Source and Wheel (universal) distribution…')
        os.system('{0} setup.py sdist bdist_wheel --universal'.format(sys.executable))

        self.status('Uploading the package to PyPi via Twine…')
        os.system('twine upload dist/*')

        self.status('Pushing git tags…')
        os.system('git tag v{0}'.format(about['__version__']))
        os.system('git push --tags')

        sys.exit()


# Where the magic happens:
setup(
    name=NAME,
    version=about['__version__'],
    description=DESCRIPTION,
    long_description=long_description,
    long_description_content_type='text/markdown',
    author=AUTHOR,
    author_email=EMAIL,
    python_requires=REQUIRES_PYTHON,
    url=URL,
    packages=find_packages(exclude=('tests', 'demo', 'data', )),
    entry_points={
        'console_scripts': ['bcipy = bcipy.main:bcipy_main'],
    },
    install_requires=REQUIRED,
    include_package_data=True,
    license='Hippocratic License 2.1',
    classifiers=[
        # Trove classifiers
        # Full list: https://pypi.python.org/pypi?%3Aaction=list_classifiers
        'License :: Other/Proprietary License',
        'Topic :: Scientific/Engineering :: Human Machine Interfaces',
        'Intended Audience :: Science/Research',
        'Intended Audience :: End Users/Desktop',
        'Intended Audience :: Developers',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
    ],
    # $ setup.py publish support.
    cmdclass={
        'upload': UploadCommand,
    },
)



from typing import List


def compute_probs_after_preview(
    inquiry: List[str], symbol_set: List[str], user_error_prob: float, proceed: bool
) -> List[float]:
    r"""
    Notation:
    - "want" = desired letter
    - "show" = shown letters
    - "button" = user's button response. "+" = user selected, "-" = user not selected
    - "past" = prior from language model and possibly previous button evidence

    Assumptions:
    1. User's errors are symmetric. p(+ | want not in show) = p(- | want in show) = user_error_prob
       This lets us use a single error rate to describe the user behavior for all cases
       (desired letter shown/not shown, response of + or -)
       The alternative would be determining one rate of accidental presses and another rate of accidental misses.

    2. Given the shown letters and user's desired letter, their response is independent of previous events.
       The user's behavior is fully determined by their desired letter and the shown letters.
       Only some edge effects such as frustration or fatigue are ignored by this.

    We're interested in the symbol_set posterior overall:

        p(want | button, show, past) = p(button | want, show, past) p(want | show, past) / p(button | show, past)
                                    # Use assumption 2.
                                    = p(button | want, show) p(want | show, past) / p(button | show, past)
                                    # Given past, show does not affect want
                                    = p(button | want, show) p(want | past) / p(button | show, past)
                                    # Will normalize over symbol_set anyway. Can ignore normalization const.
                                    \propto p(button | want, show) p(want | past)
                                    # Leave out the prior - this is handled outside by evidence fusion.
                                    \propto p(button | want, show)
                                            ~~~~~~~~~~~~~~~~~~~~~~
                                                    |
                                    These are the multiplicative updates we'll return.


    For shown letters, if the user gives a positive button response, we have terms like:

        p(want=A | +, show=ABC, past) \propto p(+ | want=A, show=ABC)
                                      \propto p(+ | want in show) = "probability of correct button press"

    And for shown letters, a negative response gives us terms like this:

        p(want=A | -, show=ABC, past) \propto p(- | want=A, show=ABC)
                                      \propto p(- | want in show) = "probability of incorrect button press"

    For unshown letter, positive button response:

        p(want=Z | +, show=ABC, past) \propto p(+ | want not in show) = "probability of incorrect button press"

    For unshown letter, negative button response:

        p(want=Z | -, show=ABC, past) \propto p(- | want not in show) = "probability of correct button press"

    Note that this simple rule, followed by normalizing the result vector, has the desired scaling property;
    with larger inquiries, the prior probability of a "+" reponse is larger.
    Thus, for a "+" response, the total probability given to the shown letters should grow as the inquiry length grows.

    >>> def example(N, L):
    ...     # Show N letters from an symbol_set of length L
    ...     probs = np.array([0.95]*N + [0.05] * (L-N))
    ...     probs /= probs.sum()
    ...     # Check the effect of a "+" response on the N shown letters
    ...     res = probs[:N].sum()
    ...     print(format(res, "0.2f"))
    >>> example(2, 10)
    0.83
    >>> example(5, 10)
    0.95
    >>> example(8, 10)
    0.99

    Args:
        inquiry - the symbols presented in the inquiry preview
        symbol_set - the full symbol_set of symbols
        proceed - whether or not the user wants to proceed with this inquiry
        user_error_prob - the probability that the user selects incorrectly during preview
    """
    if not 0 <= user_error_prob <= 1:
        raise ValueError(f"invalid user error prob: {user_error_prob}")

    if proceed:
        shown_letter_value = 1 - user_error_prob
        other_letter_value = user_error_prob
    else:
        shown_letter_value = user_error_prob
        other_letter_value = 1 - user_error_prob

    return [shown_letter_value if letter in inquiry else other_letter_value for letter in symbol_set]



from bcipy.signal.process.transform import get_default_transform, Downsample, Composition
from bcipy.signal.process.filter import filter_inquiries

__all__ = [
    "filter_inquiries",
    "get_default_transform",
    "Downsample",
    "Composition"
]



"""GUI form for collecting experimental field data."""
# pylint: disable=E0611

import sys

from typing import List

from PyQt5.QtCore import Qt
from PyQt5.QtWidgets import (
    QHBoxLayout,
    QPushButton,
    QScrollArea,
    QVBoxLayout,
    QWidget,
)

from bcipy.gui.main import (
    AlertMessageType,
    AlertMessageResponse,
    AlertResponse,
    MessageBox,
    app,
    BoolInput,
    DirectoryInput,
    FileInput,
    FloatInput,
    FormInput,
    IntegerInput,
    TextInput
)
from bcipy.helpers.load import load_experiments, load_fields
from bcipy.helpers.save import save_experiment_field_data


class ExperimentFieldCollection(QWidget):
    """Experiment Field Collection.

    Given an experiment with fields to be collected, this UI can be used to collect data in the correct format
        and require fields which are noted as such in the experiment.
    """
    field_data: List[tuple] = []
    field_inputs: List[FormInput] = []
    type_inputs = {
        'int': IntegerInput,
        'float': FloatInput,
        'bool': BoolInput,
        'filepath': FileInput,
        'directorypath': DirectoryInput
    }
    require_mark = '*'
    alert_timeout = 10
    save_data = {}

    def __init__(self, title: str, width: int, height: int, experiment_name: str, save_path: str, file_name: str):
        super().__init__()

        self.experiment_name = experiment_name
        self.experiment = load_experiments()[experiment_name]
        self.save_path = save_path
        self.file_name = file_name
        self.help_size = 12
        self.help_color = 'darkgray'
        self.width = width
        self.height = height
        self.title = title

        self.fields = load_fields()
        self.build_field_data()
        self.build_assets()
        self.do_layout()

    def do_layout(self) -> None:
        """Layout the form controls."""
        vbox = QVBoxLayout()

        # Add the controls to the grid:
        for form_input in self.field_inputs:
            vbox.addWidget(form_input)

        self.setLayout(vbox)
        self.setFixedWidth(self.width)
        self.show()

    def build_form(self) -> None:
        """Build Form.

        Loop over the field data and create UI field inputs for data collection.
        """
        for field_name, field_type, required, help_text in self.field_data:
            self.field_inputs.append(self.field_input(field_name, field_type, help_text, required))

    def field_input(self, field_name: str, field_type: str, help_tip: str, required: bool) -> FormInput:
        """Field Input.

        Construct a FormInput for the given field based on its python type and other
        attributes.
        """

        form_input = self.type_inputs.get(field_type, TextInput)

        if required:
            field_name += self.require_mark
        return form_input(
            label=field_name,
            value='',
            help_tip=help_tip,
            help_size=self.help_size,
            help_color=self.help_color)

    def build_assets(self) -> None:
        """Build Assets.

        Build any needed assets for the Experiment Field Widget. Currently, only a form is needed.
        """
        self.build_form()

    def check_input(self) -> bool:
        """Check Input.

        Ensure that any fields that require input have data!
        """
        for field in self.field_inputs:
            _input = field.value()
            name = field.label
            if self.require_mark in field.label and not _input:
                self.throw_alert_message(
                    title='BciPy Alert',
                    message=f'Required field {name.strip(self.require_mark)} must be filled out!',
                    message_type=AlertMessageType.CRIT,
                    message_response=AlertMessageResponse.OCE,
                    message_timeout=self.alert_timeout)
                return False
        return True

    def build_field_data(self) -> None:
        """Build Field Data.

        Using the fields defined in the experiment, fetch the other attributes of the field. It will be stored in
            self.field_data as a list of tuples (name, field type, required, help text).
        """
        for field in self.experiment['fields']:
            # the field name and requirement
            for name, required in field.items():
                # help text and type
                field_data = self.fields[name]
                self.field_data.append(
                    (name.title(), field_data['type'], self.map_to_bool(required['required']), field_data['help_text'])
                )

    def map_to_bool(self, string_boolean: str) -> bool:
        """Map To Boolean.

        All data is loaded from json ("true"/"false"). This method will return a python boolean (True/False).
        """
        if string_boolean == 'true':
            return True
        elif string_boolean == 'false':
            return False
        raise Exception(f'Unsupported boolean value {string_boolean}')

    def save(self) -> None:
        if self.check_input():
            self.build_save_data()
            self.write_save_data()

    def build_save_data(self) -> None:
        """Build Save Data."""
        try:
            for field in self.field_inputs:
                _input = field.cast_value()
                name = field.label.strip(self.require_mark)
                self.save_data[name] = _input
        except ValueError as e:
            self.throw_alert_message(
                title='Error',
                message=f'Error saving data. Invalid value provided. \n {e}',
                message_type=AlertMessageType.WARN,
                message_response=AlertMessageResponse.OCE,
                message_timeout=self.alert_timeout
            )

    def write_save_data(self) -> None:
        save_experiment_field_data(self.save_data, self.save_path, self.file_name)
        self.throw_alert_message(
            title='Success',
            message=(
                f'Data successfully written to: \n\n{self.save_path}/{self.file_name} \n\n\n'
                'Please wait or close this window to start the task!'),
            message_type=AlertMessageType.INFO,
            message_response=AlertMessageResponse.OCE,
            message_timeout=self.alert_timeout,
        )

    def throw_alert_message(self,
                            title: str,
                            message: str,
                            message_type: AlertMessageType = AlertMessageType.INFO,
                            message_response: AlertMessageResponse = AlertMessageResponse.OTE,
                            message_timeout: float = 0) -> MessageBox:
        """Throw Alert Message."""

        msg = MessageBox()
        msg.setWindowTitle(title)
        msg.setText(message)
        msg.setIcon(message_type.value)
        msg.setTimeout(message_timeout)

        if message_response is AlertMessageResponse.OTE:
            msg.setStandardButtons(AlertResponse.OK.value)
        elif message_response is AlertMessageResponse.OCE:
            msg.setStandardButtons(AlertResponse.OK.value | AlertResponse.CANCEL.value)

        return msg.exec_()


class MainPanel(QWidget):
    """Main GUI window.

    Parameters:
    -----------
      title: window title
      width: window width
      height: window height
      experiment_name: name of the experiment to collect field data for
      save_path: where to save the collected field data
      file_name: name of the file to write with collected field_data
    """

    def __init__(self, title: str, width: int, height: int, experiment_name: str, save_path: str, file_name: str):
        super().__init__()
        self.title = title
        self.width = width
        self.height = height

        self.form = ExperimentFieldCollection(title, width, height, experiment_name, save_path, file_name)
        self.initUI()

    def initUI(self):
        """Initialize the UI"""
        vbox = QVBoxLayout()

        self.form_panel = QScrollArea()
        self.form_panel.setVerticalScrollBarPolicy(Qt.ScrollBarAlwaysOn)
        self.form_panel.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
        self.form_panel.setWidgetResizable(True)
        self.form_panel.setFixedWidth(self.width)
        self.form_panel.setWidget(self.form)

        vbox.addWidget(self.form_panel)
        vbox.addSpacing(5)

        self.setLayout(vbox)
        self.setFixedHeight(self.height)
        self.setWindowTitle(self.title)

        control_box = QHBoxLayout()
        control_box.addStretch()
        save_button = QPushButton('Save')
        save_button.setFixedWidth(80)
        save_button.clicked.connect(self.save)
        control_box.addWidget(save_button)
        vbox.addLayout(control_box)
        self.show()

    def save(self):
        self.form.save()
        self.close()


def start_app() -> None:
    """Start Experiment Field Collection."""
    import argparse
    from bcipy.config import DEFAULT_EXPERIMENT_ID, EXPERIMENT_DATA_FILENAME
    from bcipy.helpers.validate import validate_experiment, validate_field_data_written

    parser = argparse.ArgumentParser()

    # experiment_name
    parser.add_argument('-p', '--path', default='.',
                        help='Path to save collected field data to in json format')
    parser.add_argument('-e', '--experiment', default=DEFAULT_EXPERIMENT_ID,
                        help='Select a valid experiment to run the task for this user')
    parser.add_argument('-f', '--filename', default=EXPERIMENT_DATA_FILENAME,
                        help='Provide a json filename to write the field data to. Ex, experiment_data.json')
    parser.add_argument('-v', '--validate', default=False,
                        help='Whether or not to validate the experiment before proceeding to data collection.')

    args = parser.parse_args()

    experiment_name = args.experiment
    validate = args.validate

    if validate:
        validate_experiment(experiment_name)
        print('Experiment valid!')

    save_path = args.path
    file_name = args.filename

    bcipy_gui = app(sys.argv)

    ex = MainPanel(
        title='Experiment Field Collection',
        height=250,
        width=600,
        experiment_name=experiment_name,
        save_path=save_path,
        file_name=file_name
    )
    bcipy_gui.exec_()

    if validate:
        if validate_field_data_written(save_path, file_name):
            print('Field data successfully written!')
        else:
            raise Exception(f'Field data not written to {save_path}/{file_name}')

    sys.exit()


if __name__ == '__main__':
    start_app()



from psychopy import visual
from bcipy.display.paradigm.rsvp.display import RSVPDisplay, BCIPY_LOGO_PATH
from bcipy.helpers.symbols import SPACE_CHAR
from bcipy.helpers.stimuli import resize_image

"""Note:

RSVP Tasks are RSVPDisplay objects with different structure. They share
the tasks and the essential elements and stimuli. However, layout, length of
stimuli list, update procedures and colors are different. Therefore each
mode should be separated from each other carefully.
Functions:
    update_task_state: update task information of the module
"""


class CopyPhraseDisplay(RSVPDisplay):
    """ Copy Phrase display object of RSVP

        Custom attributes:
            static_task_text(str): target text for the user to attempt to spell
            static_task_color(str): target text color for the user to attempt to spell
    """

    def __init__(
            self,
            window,
            clock,
            experiment_clock,
            stimuli,
            task_bar,
            info,
            starting_spelled_text='',
            trigger_type='image',
            space_char=SPACE_CHAR,
            preview_inquiry=None,
            full_screen=False):
        """ Initializes Copy Phrase Task Objects """
        self.starting_spelled_text = starting_spelled_text

        super().__init__(window,
                         clock,
                         experiment_clock,
                         stimuli,
                         task_bar,
                         info,
                         trigger_type=trigger_type,
                         space_char=space_char,
                         preview_inquiry=preview_inquiry,
                         full_screen=full_screen)

    def wait_screen(self, message: str, message_color: str) -> None:
        """Wait Screen.

        Args:
            message(string): message to be displayed while waiting
            message_color(string): color of the message to be displayed
        """

        self.draw_static()

        # Construct the wait message
        wait_message = visual.TextStim(win=self.window,
                                       font=self.stimuli_font,
                                       text=message,
                                       height=.1,
                                       color=message_color,
                                       pos=(0, -.55),
                                       wrapWidth=2,
                                       colorSpace='rgb',
                                       opacity=1,
                                       depth=-6.0)

        # try adding the BciPy logo to the wait screen
        try:
            wait_logo = visual.ImageStim(
                self.window,
                image=BCIPY_LOGO_PATH,
                pos=(0, 0),
                mask=None,
                ori=0.0)
            wait_logo.size = resize_image(
                BCIPY_LOGO_PATH,
                self.window.size,
                1)
            wait_logo.draw()

        except Exception as e:
            self.logger.exception(f'Cannot load logo image from path=[{BCIPY_LOGO_PATH}]')
            raise e

        # Draw and flip the screen.
        wait_message.draw()
        self.window.flip()



import logging

from typing import Optional
import numpy as np
from sklearn.decomposition import PCA


class ChannelWisePrincipalComponentAnalysis:
    """Creates a PCA object for each channel.

    Args:
        n_components(int or float): Number of components in PCA
            - If 0 < n_components < 1, keeps enough components to preserve the specified fraction of variance.
            - If int, keeps the first n_components components.
            - If None, keeps all components

            NOTE - Keeping more components (or using a float closer to 1) preserves more structure;
            keeping fewer (or using a float closer to 0) achieves more compression and thus more speedup for
            downstream steps. The tradeoff between compression and performance is fundamentally arbitrary.
            To choose number of components, try plotting the explained_variance_ratio_ attribute of the PCA object
            on a representative dataset, and check the effect on runtime and task performance.

        random_state(int): Random state seed
        num_ch(int): number of channels in expected data
    """

    def __init__(self, n_components: Optional[float] = None, random_state: Optional[int] = None, num_ch: int = 1):
        self.num_ch = num_ch
        self.list_pca = [PCA(n_components=n_components, random_state=random_state) for _ in range(self.num_ch)]
        self.logger = logging.getLogger(__name__)
        self.logger.info(f"PCA. n_components={n_components}, random_state={random_state}, num_ch={num_ch}")

    def fit(self, x: np.ndarray, y: np.ndarray = None) -> None:
        """Fit PCA to each channel of data.

        Args:
            x(ndarray[float]): C x N x k data array
            y(ndarray[int]): N x k observation (class) array
                N is number of samples k is dimensionality of features
                C is number of channels
        """
        for i in range(self.num_ch):
            self.list_pca[i].fit(x[i, :, :], y)

    def transform(self, x: np.ndarray, y: np.ndarray = None) -> np.ndarray:
        """Apply fitted PCA to each channel.

        Args:
            x: data, with shape (channels, items, samples)
            y: labels (ignored)

        Returns:
            reduced dim data, shape (items, K), where K is the size after reducing and concatenating all channels
        """
        f_vector = []
        for i in range(self.num_ch):
            f_vector.append(self.list_pca[i].transform(x[i, :, :]))
        return np.concatenate(f_vector, axis=1)

    def fit_transform(self, x: np.ndarray, y: np.ndarray = None) -> np.ndarray:
        self.fit(x)
        return self.transform(x)


class MockPCA:
    """Mock PCA object for testing purposes."""

    def __init__(self, *args, **kw):
        pass

    def fit(self, *args, **kw):
        pass

    def transform(self, x: np.ndarray, y: np.ndarray = None, var_tol: Optional[float] = None) -> np.ndarray:
        """Mock transform method for testing purposes.

        Args:
            x: data, with shape (channels, items, samples)
            y: labels (ignored)
            var_tol: variance tolerance (ignored)

        Returns:
            reduced dim data, shape (items, K), where K is the size after reducing and concatenating all channels
        """
        x = x.transpose([1, 0, 2])
        x = x.reshape([x.shape[0], -1])
        return x

    def fit_transform(self, x: np.ndarray, y: np.ndarray = None, var_tol: Optional[float] = None) -> np.ndarray:
        self.fit(x)
        return self.transform(x, y, var_tol)



"""Tests for data conversion related functionality."""
import os
import shutil
import tempfile
from typing import Tuple, Union
import unittest
import warnings

from pathlib import Path

from mockito import mock, when, unstub, verify, verifyNoMoreInteractions, any as any_value

from bcipy.config import DEFAULT_ENCODING, RAW_DATA_FILENAME, TRIGGER_FILENAME, DEFAULT_PARAMETER_FILENAME
from bcipy.helpers import convert
from bcipy.helpers.convert import (
    archive_list,
    compress,
    convert_to_bdf,
    convert_to_edf,
    convert_to_mne,
    decompress,
    pyedf_convert,
    tobii_to_norm,
    norm_to_tobii,
)
from bcipy.helpers.parameters import Parameters
from bcipy.helpers.raw_data import sample_data, write, RawData
from bcipy.helpers.triggers import MOCK_TRIGGER_DATA
from bcipy.signal.generator.generator import gen_random_data

from mne.io import read_raw_edf, read_raw_bdf


def create_bcipy_session_artifacts(
        write_dir: str,
        channels: Union[int, list] = 3,
        sample_rate: int = 300,
        samples: int = 5000,
        filter_settings: dict = {
            'filter_low': 0.5,
            'filter_high': 30,
            'filter_order': 5,
            'notch_filter_frequency': 60,
            'down_sampling_rate': 3,
            'static_trigger_offset': 0.0
        },
) -> Tuple[str, RawData, Parameters]:
    """Write BciPy session artifacts to a temporary directory.

    This includes a raw data file, trigger file, and a parameters file.
    """
    trg_data = MOCK_TRIGGER_DATA
    if isinstance(channels, int):
        channels = [f'ch{i}' for i in range(channels)]
    data = sample_data(ch_names=channels, sample_rate=sample_rate, rows=samples)
    with open(Path(write_dir, TRIGGER_FILENAME), 'w', encoding=DEFAULT_ENCODING) as trg_file:
        trg_file.write(trg_data)

    write(data, Path(write_dir, f'{RAW_DATA_FILENAME}.csv'))

    params = Parameters.from_cast_values(raw_data_name=f'{RAW_DATA_FILENAME}.csv',
                                         trigger_file_name=TRIGGER_FILENAME,
                                         preview_inquiry_length=0.5,
                                         time_fixation=0.5,
                                         time_prompt=0.5,
                                         time_flash=0.5,
                                         # define filter settings
                                         static_trigger_offset=filter_settings['static_trigger_offset'],
                                         down_sampling_rate=filter_settings['down_sampling_rate'],
                                         notch_filter_frequency=filter_settings['notch_filter_frequency'],
                                         filter_high=filter_settings['filter_high'],
                                         filter_low=filter_settings['filter_low'],
                                         filter_order=filter_settings['filter_order'])
    params.save(write_dir, DEFAULT_PARAMETER_FILENAME)
    return trg_data, data, params


class TestEDFConvert(unittest.TestCase):
    """Tests for EDF data format conversions."""

    def setUp(self):
        self.temp_dir = tempfile.mkdtemp()
        _, data, _ = create_bcipy_session_artifacts(self.temp_dir)
        self.channels = data.channels

    def tearDown(self):
        shutil.rmtree(self.temp_dir)
        verifyNoMoreInteractions()
        unstub()

    def test_convert_to_edf_defaults(self):
        """Test default behavior"""
        path = convert_to_edf(self.temp_dir)
        self.assertTrue(os.path.exists(path))

        with warnings.catch_warnings():
            warnings.simplefilter('ignore')
            edf = read_raw_edf(path, preload=True)

        self.assertTrue(len(edf.get_data()) > 0)

        for ch_name in self.channels:
            self.assertTrue(ch_name in edf.ch_names)

    def test_convert_to_edf_overwrite_false(self):
        """Test overwriting fails"""

        convert_to_edf(self.temp_dir)
        with self.assertRaises(OSError):
            convert_to_edf(self.temp_dir, overwrite=False)

    def test_convert_to_edf_overwrite_true(self):
        """Test that overwriting can be configured"""

        convert_to_edf(self.temp_dir)
        convert_to_edf(self.temp_dir, overwrite=True)

    def test_convert_to_edf_with_custom_path(self):
        """Test creating the EDF using a custom edf path"""
        path = convert_to_edf(self.temp_dir,
                              edf_path=str(Path(self.temp_dir, 'mydata.edf').resolve()))

        self.assertEqual(Path(path).name, 'mydata.edf')

    def test_convert_to_edf_raises_value_error_with_invalid_edfpath(self):
        """Test that an value error is raised when the edf path is invalid"""
        edf_path = 'invalid.pdf'
        with self.assertRaises(ValueError):
            convert_to_edf(self.temp_dir, edf_path=edf_path)

    def test_convert_to_edf_with_write_targetness(self):
        """Test creating the EDF using targetness for event annotations"""
        path = convert_to_edf(self.temp_dir,
                              write_targetness=True)

        with warnings.catch_warnings():
            warnings.simplefilter('ignore')
            edf = read_raw_edf(path, preload=True)

        self.assertIn('target', edf.annotations.description)
        self.assertIn('nontarget', edf.annotations.description)

    def test_convert_to_edf_without_write_targetness(self):
        """Test creating the EDF with labels as event annotations"""
        path = convert_to_edf(self.temp_dir,
                              write_targetness=False)

        with warnings.catch_warnings():
            warnings.simplefilter('ignore')
            edf = read_raw_edf(path, preload=True)

        self.assertNotIn('target', edf.annotations.description)
        self.assertNotIn('nontarget', edf.annotations.description)
        self.assertIn('+', edf.annotations.description)

    def test_convert_to_edf_with_pre_filter(self):
        """Test creating the EDF with pre-filtering"""

        data = mock()
        channels = mock()
        fs = 100
        events = mock()
        annotation_channels = 1
        return_path = mock()
        when(convert).pyedf_convert(
            any_value(str),
            write_targetness=any_value(bool),
            use_event_durations=any_value(bool),
            remove_pre_fixation=any_value(bool),
            pre_filter=True,
        ).thenReturn((
            data,
            channels,
            fs,
            events,
            annotation_channels,
            'filter_settings'
        ))
        when(convert).write_pyedf(
            any_value(str),
            data,
            channels,
            fs,
            events,
            any_value(int),
            any_value(bool),
            annotation_channels,
            any_value(str)).thenReturn(return_path)
        path = convert_to_edf(self.temp_dir,
                              pre_filter=True)

        verify(convert, times=1).pyedf_convert(
            any_value(str),
            write_targetness=any_value(bool),
            use_event_durations=any_value(bool),
            remove_pre_fixation=any_value(bool),
            pre_filter=True,
        )
        verify(convert, times=1).write_pyedf(
            any_value(str),
            data,
            channels,
            fs,
            events,
            any_value(int),
            any_value(bool),
            annotation_channels,
            any_value(str))
        self.assertEqual(path, return_path)


class TestBDFConvert(unittest.TestCase):
    """Tests for BDF data format conversions."""

    def setUp(self):
        """Override; set up the needed path for load functions."""

        self.temp_dir = tempfile.mkdtemp()
        _, data, _ = create_bcipy_session_artifacts(self.temp_dir)
        self.channels = data.channels

    def tearDown(self):
        """Override"""
        shutil.rmtree(self.temp_dir)
        verifyNoMoreInteractions()
        unstub()

    def test_convert_to_bdf_defaults(self):
        """Test default convert to bdf behavior"""
        path = convert_to_bdf(self.temp_dir)
        self.assertTrue(os.path.exists(path))

        with warnings.catch_warnings():
            warnings.simplefilter('ignore')
            bdf = read_raw_bdf(path, preload=True)

        self.assertTrue(len(bdf.get_data()) > 0)

        for ch_name in self.channels:
            self.assertTrue(ch_name in bdf.ch_names)

    def test_convert_to_bdf_overwrite_false(self):
        """Test overwriting fails if not configured"""

        convert_to_bdf(self.temp_dir)
        with self.assertRaises(OSError):
            convert_to_bdf(self.temp_dir, overwrite=False)

    def test_convert_to_bdf_overwrite_true(self):
        """Test that overwriting can be configured"""

        convert_to_bdf(self.temp_dir)
        convert_to_bdf(self.temp_dir, overwrite=True)

    def test_convert_to_bdf_with_custom_path(self):
        """Test creating the EDF using a custom edf path"""
        path = convert_to_bdf(self.temp_dir,
                              bdf_path=str(Path(self.temp_dir, 'mydata.bdf').resolve()))

        self.assertEqual(Path(path).name, 'mydata.bdf')

    def test_convert_to_bdf_raises_value_error_with_invalid_edfpath(self):
        """Test that a value error is raised with the bdf path is invalid"""
        bdf_path = 'invalid.pdf'
        with self.assertRaises(ValueError):
            convert_to_bdf(self.temp_dir, bdf_path=bdf_path)

    def test_convert_to_bdf_with_write_targetness(self):
        """Test creating the EDF using targetness for event annotations"""
        path = convert_to_bdf(self.temp_dir,
                              write_targetness=True)

        with warnings.catch_warnings():
            warnings.simplefilter('ignore')
            edf = read_raw_bdf(path, preload=True)

        self.assertIn('target', edf.annotations.description)
        self.assertIn('nontarget', edf.annotations.description)

    def test_convert_to_bdf_without_write_targetness(self):
        """Test creating the EDF with labels as event annotations"""
        path = convert_to_bdf(self.temp_dir,
                              write_targetness=False)

        with warnings.catch_warnings():
            warnings.simplefilter('ignore')
            edf = read_raw_bdf(path, preload=True)

        self.assertNotIn('target', edf.annotations.description)
        self.assertNotIn('nontarget', edf.annotations.description)
        self.assertIn('+', edf.annotations.description)

    def test_convert_to_bdf_with_pre_filter(self):
        """Test creating the BDF with pre-filtering"""

        data = mock()
        channels = mock()
        fs = 100
        events = mock()
        annotation_channels = 1
        return_path = mock()
        when(convert).pyedf_convert(
            any_value(str),
            write_targetness=any_value(bool),
            use_event_durations=any_value(bool),
            remove_pre_fixation=any_value(bool),
            pre_filter=True,
        ).thenReturn((
            data,
            channels,
            fs,
            events,
            annotation_channels,
            'filter_settings'
        ))
        when(convert).write_pyedf(
            any_value(str),
            data,
            channels,
            fs,
            events,
            any_value(int),
            any_value(bool),
            annotation_channels,
            any_value(str)).thenReturn(return_path)
        path = convert_to_bdf(self.temp_dir,
                              pre_filter=True)

        verify(convert, times=1).pyedf_convert(
            any_value(str),
            write_targetness=any_value(bool),
            use_event_durations=any_value(bool),
            remove_pre_fixation=any_value(bool),
            pre_filter=True,
        )
        verify(convert, times=1).write_pyedf(
            any_value(str),
            data,
            channels,
            fs,
            events,
            any_value(int),
            any_value(bool),
            annotation_channels,
            any_value(str))
        self.assertEqual(path, return_path)


class TestPyedfconvert(unittest.TestCase):

    def setUp(self) -> None:
        self.temp_dir = tempfile.mkdtemp()
        self.sample_rate = 300
        self.filter_settings = {
            'filter_low': 0.5,
            'filter_high': 30,
            'filter_order': 5,
            'notch_filter_frequency': 60,
            'down_sampling_rate': 3,
            'static_trigger_offset': 0.0
        }
        create_bcipy_session_artifacts(
            self.temp_dir,
            sample_rate=self.sample_rate,
            filter_settings=self.filter_settings,
            samples=10000)  # sample number determines the annotation count, set it high here!

    def tearDown(self) -> None:
        shutil.rmtree(self.temp_dir)
        verifyNoMoreInteractions()
        unstub()

    def test_pyedf_convert_defaults(self):
        """Test the pyedf_convert function"""
        data, channels, fs, events, annotation_channels, pre_filter = pyedf_convert(
            self.temp_dir)

        self.assertTrue(len(data) > 0)
        self.assertTrue(len(channels) > 0)
        self.assertEqual(fs, self.sample_rate)
        self.assertTrue(len(events) > 2)
        # see the validate_annotations function for details on how this is calculated
        self.assertEqual(annotation_channels, 1)
        self.assertFalse(pre_filter)

    def test_pyedf_convert_with_pre_filter(self):
        """Test the pyedf_convert function with pre-filtering"""
        data, channels, fs, events, annotation_channels, pre_filter = pyedf_convert(
            self.temp_dir, pre_filter=True)

        self.assertTrue(len(data) > 0)
        self.assertTrue(len(channels) > 0)
        self.assertEqual(fs, self.sample_rate / self.filter_settings['down_sampling_rate'])
        self.assertTrue(len(events) > 2)
        # see the validate_annotations function for details on how this is calculated
        #  with less samples, the number of annotation channels increases
        self.assertEqual(annotation_channels, 2)
        self.assertIsInstance(pre_filter, str)

    def test_pyedf_convert_with_write_targetness(self):
        """Test the pyedf_convert function with targetness for event annotations"""
        data, channels, fs, events, annotation_channels, _ = pyedf_convert(
            self.temp_dir, write_targetness=True)

        self.assertTrue(len(data) > 0)
        self.assertTrue(len(channels) > 0)
        self.assertEqual(fs, self.sample_rate)
        self.assertTrue(len(events) > 2)
        # see the validate_annotations function for details on how this is calculated
        self.assertEqual(annotation_channels, 1)
        for event in events:
            _, _, label = event
            # in this casae label and targetness should not be the same
            self.assertTrue(label in ['target', 'nontarget'])

    def test_pyedf_convert_without_write_targetness(self):
        """Test the pyedf_convert function with labels as event annotations"""
        data, channels, fs, events, annotation_channels, _ = pyedf_convert(
            self.temp_dir, write_targetness=False)

        self.assertTrue(len(data) > 0)
        self.assertTrue(len(channels) > 0)
        self.assertEqual(fs, self.sample_rate)
        self.assertTrue(len(events) > 2)
        # see the validate_annotations function for details on how this is calculated
        self.assertEqual(annotation_channels, 1)
        for event in events:
            _, _, label = event
            self.assertTrue(label not in ['target', 'nontarget'])

    def test_pyedf_convert_with_use_event_durations(self):
        """Test the pyedf_convert function with event durations"""
        data, channels, fs, events, annotation_channels, _ = pyedf_convert(
            self.temp_dir, use_event_durations=True)

        self.assertTrue(len(data) > 0)
        self.assertTrue(len(channels) > 0)
        self.assertEqual(fs, self.sample_rate)
        self.assertTrue(len(events) > 2)
        # see the validate_annotations function for details on how this is calculated
        self.assertEqual(annotation_channels, 1)
        for _, duration, _ in events:
            self.assertTrue(duration > 0)

    def test_pyedf_convert_without_use_event_durations(self):
        """Test the pyedf_convert function without event durations"""
        data, channels, fs, events, annotation_channels, _ = pyedf_convert(
            self.temp_dir, use_event_durations=False)

        self.assertTrue(len(data) > 0)
        self.assertTrue(len(channels) > 0)
        self.assertEqual(fs, self.sample_rate)
        self.assertTrue(len(events) > 2)
        # see the validate_annotations function for details on how this is calculated
        self.assertEqual(annotation_channels, 1)
        for _, duration, _ in events:
            self.assertEqual(duration, 0)

    def test_pyedf_convert_with_remove_pre_fixation(self):
        """Test the pyedf_convert function with pre-fixation removal"""
        data, channels, fs, events, annotation_channels, _ = pyedf_convert(
            self.temp_dir, remove_pre_fixation=True)

        self.assertTrue(len(data) > 0)
        self.assertTrue(len(channels) > 0)
        self.assertEqual(fs, self.sample_rate)
        self.assertTrue(len(events) > 2)
        # see the validate_annotations function for details on how this is calculated
        self.assertEqual(annotation_channels, 1)
        for event in events:
            self.assertNotIn('fixation', event)
            self.assertNotIn('prompt', event)


class TestMNEConvert(unittest.TestCase):
    """Test the convert_to_mne function, which converts bcipy RawData into an mne format"""

    def setUp(self):
        """Set up the test case with a temporary directory and sample data"""
        self.temp_dir = tempfile.mkdtemp()
        self.sample_rate = 300
        self.filter_settings = {
            'filter_low': 0.5,
            'filter_high': 30,
            'filter_order': 5,
            'notch_filter_frequency': 60,
            'down_sampling_rate': 3,
            'static_trigger_offset': 0.0
        }
        self.channels = ['timestamp', 'O1', 'O2', 'Pz']
        self.raw_data = RawData('SampleDevice', self.sample_rate, self.channels)
        # generate 100 random samples of data
        for _ in range(0, 100):
            channel_data = gen_random_data(low=-1000,
                                           high=1000,
                                           channel_count=len(self.channels))
            self.raw_data.append(channel_data)

    def tearDown(self):
        """Remove the temporary directory and its contents after each test"""
        shutil.rmtree(self.temp_dir)

    def test_convert_to_mne_defaults(self):
        """Test the convert_to_mne function with default parameters"""
        data = convert_to_mne(self.raw_data)

        self.assertTrue(len(data) > 0)
        self.assertEqual(data.ch_names, self.channels[1:])
        self.assertEqual(data.info['sfreq'], self.sample_rate)

    def test_convert_to_mne_with_channel_map(self):
        """Test the convert_to_mne function with channel mapping"""
        # here we know only three channels are generated, using the channel map let's only use the last one
        channel_map = [0, 0, 1]
        data = convert_to_mne(self.raw_data, channel_map=channel_map)

        self.assertTrue(len(data) > 0)
        self.assertTrue(len(data.ch_names) == 1)  # this is the main assertion!
        self.assertEqual(data.info['sfreq'], self.sample_rate)

    def test_convert_to_mne_with_channel_types(self):
        """Test the convert_to_mne function with channel types"""
        channel_types = ['eeg', 'eeg', 'seeg']
        data = convert_to_mne(self.raw_data, channel_types=channel_types)

        self.assertTrue(len(data) > 0)
        self.assertEqual(data.ch_names, self.channels[1:])
        self.assertEqual(data.info['sfreq'], self.sample_rate)
        self.assertTrue(data.get_channel_types()[2] == 'seeg')

    def test_convert_to_mne_with_transform(self):
        """Test the convert_to_mne function with a transform"""
        multiplier = 2

        def transform(x, fs):
            return x * multiplier, fs

        data = convert_to_mne(self.raw_data, transform=transform)

        self.assertTrue(len(data) > 0)
        self.assertEqual(data.ch_names, self.channels[1:])
        self.assertEqual(data.info['sfreq'], self.sample_rate)

        # apply the transform to the first data point and compare to data returned
        expected_first_data_point = self.raw_data.channel_data[0][0] * multiplier
        self.assertTrue(data.get_data()[0][0] == expected_first_data_point)

    def test_convert_to_mne_with_mv_conversion(self):
        """Test the convert_to_mne function with a mv conversion"""
        data = convert_to_mne(self.raw_data, volts=False)

        self.assertTrue(len(data) > 0)
        self.assertEqual(data.ch_names, self.channels[1:])
        self.assertEqual(data.info['sfreq'], self.sample_rate)

        # apply the transform to the first data point and compare to data returned
        expected_first_data_point = self.raw_data.channel_data[0][0] * 1e-6
        self.assertTrue(data.get_data()[0][0] == expected_first_data_point)

    def test_convert_to_mne_with_custom_montage(self):
        """Test the convert_to_mne function with a custom montage"""

        # see https://mne.tools/stable/auto_tutorials/intro/40_sensor_locations.html
        # for more information on montages and available defaults
        montage_type = 'biosemi64'
        data = convert_to_mne(self.raw_data, montage=montage_type)

        self.assertTrue(len(data) > 0)
        self.assertEqual(data.ch_names, self.channels[1:])
        self.assertEqual(data.info['sfreq'], self.sample_rate)


class TestCompressionSupport(unittest.TestCase):

    def setUp(self):
        self.dir_name = 'test/'
        self.tar_file_name = 'test_file'
        self.tar_file_full_name = f'{self.tar_file_name}.tar.gz'
        self.test_file_name = 'test.text'
        with open(self.test_file_name, 'w', encoding=DEFAULT_ENCODING) as fp:
            pass

    def tearDown(self):
        os.remove(self.test_file_name)

        if os.path.exists(self.tar_file_full_name):
            os.remove(self.tar_file_full_name)

    def test_compression_writes_tar_gz_no_extension(self):
        # Test with no extension on tar name
        compress(self.tar_file_name, [self.test_file_name])
        # Assert correct file was written
        self.assertTrue(os.path.exists(self.tar_file_full_name))

    def test_compression_writes_tar_gz_with_extension(self):
        # Test with extension on tar name
        compress(self.tar_file_full_name, [self.test_file_name])
        # Assert correct file was written
        self.assertTrue(os.path.exists(self.tar_file_full_name))

    def test_decompression_extracts_file_no_extension(self):
        # Test with no extension on tar name
        compress(self.tar_file_name, [self.test_file_name])
        decompress(self.tar_file_name, ".")
        self.assertTrue(os.path.exists(self.test_file_name))

    def test_decompression_extracts_file_with_extension(self):
        # Test with extension on tar name
        compress(self.tar_file_name, [self.test_file_name])
        decompress(self.tar_file_full_name, ".")
        self.assertTrue(os.path.exists(self.test_file_name))

    def test_file_not_found_error_thrown_on_compression(self):
        garbage_name = 'not_possible_to_exist.biz'
        with self.assertRaises(FileNotFoundError):
            compress(self.tar_file_name, [garbage_name])

    def test_file_list_returns_compressed_file_name_no_extension(self):
        # Test with no extension on tar name
        compress(self.tar_file_name, [self.test_file_name])
        tar_list = archive_list(self.tar_file_name)
        self.assertTrue(tar_list[0] == self.test_file_name)

    def test_file_list_returns_compressed_file_name_with_extension(self):
        # Test with extension on tar name
        compress(self.tar_file_name, [self.test_file_name])
        tar_list = archive_list(self.tar_file_full_name)
        self.assertTrue(tar_list[0] == self.test_file_name)


class TestConvertTobii(unittest.TestCase):

    def test_tobii_to_norm(self):
        """Test the tobii_to_norm function"""
        tobii_data = (0.5, 0.5)  # center of screen in tobii coordinates
        excepted_norm_data = (0, 0)  # center of screen in norm coordinates
        norm_data = tobii_to_norm(tobii_data)
        self.assertEqual(norm_data, excepted_norm_data)

        tobii_data = (0, 0)  # top left of screen in tobii coordinates
        excepted_norm_data = (-1, 1)  # top left of screen in norm coordinates
        norm_data = tobii_to_norm(tobii_data)
        self.assertEqual(norm_data, excepted_norm_data)

        tobii_data = (1, 1)  # bottom right of screen in tobii coordinates
        excepted_norm_data = (1, -1)  # bottom right of screen in norm coordinates
        norm_data = tobii_to_norm(tobii_data)
        self.assertEqual(norm_data, excepted_norm_data)

    def test_tobii_to_norm_raises_error_with_invalid_units(self):
        """Test the tobii_to_norm function raises an error with invalid units"""
        tobii_data = (-1, 1)  # invalid tobii coordinates
        with self.assertRaises(AssertionError):
            tobii_to_norm(tobii_data)

        tobii_data = (1, 11)  # invalid tobii coordinates

        with self.assertRaises(AssertionError):
            tobii_to_norm(tobii_data)

    def test_norm_to_tobii(self):
        """Test the norm_to_tobii function"""
        norm_data = (0, 0)  # center of screen in norm coordinates
        excepted_tobii_data = (0.5, 0.5)  # center of screen in tobii coordinates
        tobii_data = norm_to_tobii(norm_data)
        self.assertEqual(tobii_data, excepted_tobii_data)

        norm_data = (-1, 1)  # top left of screen in norm coordinates
        excepted_tobii_data = (0, 0)  # top left of screen in tobii coordinates
        tobii_data = norm_to_tobii(norm_data)
        self.assertEqual(tobii_data, excepted_tobii_data)

        norm_data = (1, -1)  # bottom right of screen in norm coordinates
        excepted_tobii_data = (1, 1)  # bottom right of screen in tobii coordinates
        tobii_data = norm_to_tobii(norm_data)
        self.assertEqual(tobii_data, excepted_tobii_data)

    def test_norm_to_tobii_raises_error_with_invalid_units(self):
        """Test the norm_to_tobii function raises an error with invalid units"""
        norm_data = (-1.1, 1)
        with self.assertRaises(AssertionError):
            norm_to_tobii(norm_data)

        norm_data = (1, 1.1)
        with self.assertRaises(AssertionError):
            norm_to_tobii(norm_data)


if __name__ == '__main__':
    unittest.main()



"""Streams file data for the viewer"""
import logging
import time
from bcipy.acquisition.util import StoppableThread
from bcipy.helpers.raw_data import RawDataReader
log = logging.getLogger(__name__)


class FileStreamer(StoppableThread):
    """Process that continuously reads data from a raw_data.csv file and
    publishes it to a Queue at the frequency specified in the metadata.

    Parameters
    ----------
        data_file: str
            raw data file from which to read
        data_queue : Queue
            Data will be written to the queue as it is acquired.
    """

    def __init__(self, data_file, data_queue):
        super(FileStreamer, self).__init__()
        self.data_file = data_file
        self.data_queue = data_queue

    def run(self):
        log.debug("Starting raw_data file streamer")

        with RawDataReader(self.data_file, convert_data=True) as reader:
            fs = reader.sample_rate
            log.debug(f"Publishing data at sample rate {fs} hz")

            # publish data
            for data in reader:
                if not self.running():
                    break
                self.data_queue.put(data)
                time.sleep(1 / fs)
            log.debug("Stopping raw_data file streamer")



from itertools import cycle
from bcipy.task import Task
from bcipy.task.paradigm.rsvp.calibration.calibration import RSVPCalibrationTask
from bcipy.helpers.stimuli import PhotoDiodeStimuli, get_fixation, jittered_timing


class RSVPTimingVerificationCalibration(Task):
    """RSVP Calibration Task.

    This task is used for verifying display timing by alternating solid and empty boxes. These
        stimuli can be used with a photodiode to ensure accurate presentations.

    Input:
        win (PsychoPy Display Object)
        daq (Data Acquisition Object)
        parameters (Dictionary)
        file_save (String)

    Output:
        file_save (String)
    """
    TASK_NAME = 'RSVP Timing Verification Task'

    def __init__(self, win, daq, parameters, file_save):
        super(RSVPTimingVerificationCalibration, self).__init__()
        parameters['stim_height'] = 0.8
        parameters['stim_pos_y'] = 0.0
        self.stimuli = PhotoDiodeStimuli.list()
        self._task = RSVPCalibrationTask(win, daq, parameters, file_save)
        self._task.generate_stimuli = self.generate_stimuli

    def generate_stimuli(self):
        """Generates the inquiries to be presented.
        Returns:
        --------
            tuple(
                samples[list[list[str]]]: list of inquiries
                timing(list[list[float]]): list of timings
                color(list(list[str])): list of colors)
        """
        samples, times, colors = [], [], []

        # alternate between solid and empty boxes
        letters = cycle(self.stimuli)
        time_prompt, time_fixation, time_stim = self._task.timing
        color_target, color_fixation, color_stim = self._task.color
        inq_len = self._task.stim_length

        stim_timing = jittered_timing(time_stim, self._task.jitter, inq_len)
        target = next(letters)
        fixation = get_fixation(is_txt=True)

        inq_stim = [target, fixation, *[next(letters) for _ in range(inq_len)]]
        inq_times = [time_prompt, time_fixation] + stim_timing
        inq_colors = [color_target, color_fixation, *[color_stim for _ in range(inq_len)]]

        for _ in range(self._task.stim_number):
            samples.append(inq_stim)
            times.append(inq_times)
            colors.append(inq_colors)

        return (samples, times, colors)

    def execute(self):
        self.logger.debug(f'Starting {self.name()}!')
        self._task.execute()

    @classmethod
    def label(cls):
        return RSVPTimingVerificationCalibration.TASK_NAME

    def name(self):
        return RSVPTimingVerificationCalibration.TASK_NAME



"""Code for constructing and executing registered tasks"""
from bcipy.task.paradigm.rsvp.calibration.calibration import RSVPCalibrationTask
from bcipy.task.paradigm.rsvp.copy_phrase import RSVPCopyPhraseTask
from bcipy.task.paradigm.rsvp.calibration.timing_verification import RSVPTimingVerificationCalibration

from bcipy.task.paradigm.matrix.calibration import MatrixCalibrationTask
from bcipy.task.paradigm.matrix.timing_verification import MatrixTimingVerificationCalibration
from bcipy.task.paradigm.matrix.copy_phrase import MatrixCopyPhraseTask

from bcipy.task import Task
from bcipy.task.exceptions import TaskRegistryException
from bcipy.task.task_registry import TaskType


def make_task(display_window, daq, task, parameters, file_save,
              signal_model=None, language_model=None, fake=True) -> Task:
    """Creates a Task based on the provided parameters.

    Parameters:
    -----------
        display_window: psychopy Window
        daq: DataAcquisitionClient
        task: TaskType
        parameters: dict
        file_save: str - path to file in which to save data
        signal_model
        language_model - language model
        fake: boolean - true if eeg stream is randomly generated
    Returns:
    --------
        Task instance
    """

    # NORMAL RSVP MODES
    if task is TaskType.RSVP_CALIBRATION:
        return RSVPCalibrationTask(
            display_window, daq, parameters, file_save)

    if task is TaskType.RSVP_COPY_PHRASE:
        return RSVPCopyPhraseTask(
            display_window, daq, parameters, file_save, signal_model,
            language_model, fake=fake)

    if task is TaskType.RSVP_TIMING_VERIFICATION_CALIBRATION:
        return RSVPTimingVerificationCalibration(display_window, daq, parameters, file_save)

    if task is TaskType.MATRIX_CALIBRATION:
        return MatrixCalibrationTask(
            display_window, daq, parameters, file_save
        )

    if task is TaskType.MATRIX_TIMING_VERIFICATION_CALIBRATION:
        return MatrixTimingVerificationCalibration(display_window, daq, parameters, file_save)

    if task is TaskType.MATRIX_COPY_PHRASE:
        return MatrixCopyPhraseTask(
            display_window, daq, parameters, file_save, signal_model,
            language_model, fake=fake)
    raise TaskRegistryException(
        'The provided experiment type is not registered.')


def start_task(display_window, daq, task, parameters, file_save,
               signal_model=None, language_model=None, fake=True):
    """Creates a Task and starts execution."""
    task = make_task(display_window, daq, task, parameters, file_save,
                     signal_model, language_model, fake)
    task.execute()



"""Tests for clock-related functionality"""
import time
import unittest

from pylsl import local_clock

from bcipy.helpers.clock import Clock


class CustomTime:
    """Callable object that increments a counter on each call."""

    def __init__(self, start: int = 0):
        self.val = start - 1

    def __call__(self):
        self.val = self.val + 1
        return self.val


class TestClock(unittest.TestCase):
    """Tests for monotonic clock."""

    def test_monotonic_clock(self):
        """Test default clock behavior"""
        clock = Clock()
        time1 = clock.getTime()
        self.assertTrue(time1 > 0)

        clock.reset()
        time2 = clock.getTime()
        self.assertTrue(time2 > time1, "Reset should have no effect")

        # Delta value is machine dependent. It is likely closer than this.
        self.assertAlmostEqual(
            clock.getTime(),
            local_clock(),
            delta=4,
            msg="Default implementation uses LSL local clock.")

    def test_start_from_zero(self):
        """Test that clock can be started from zero."""
        clock = Clock(start_at_zero=True)
        self.assertEqual(0, round(clock.getTime()))
        time.sleep(0.6)

        self.assertEqual(1, round(clock.getTime()))
        clock.reset()
        self.assertEqual(0, round(clock.getTime()))

    def test_clock_customization(self):
        """Test that clock can be customized"""

        clock = Clock(get_time=CustomTime(start=0))
        self.assertEqual(1, clock.getTime())
        self.assertEqual(2, clock.getTime())



import os
import shutil
import unittest
import tempfile

from bcipy.config import DEFAULT_PARAMETERS_PATH, DEFAULT_PARAMETER_FILENAME, DEFAULT_EXPERIMENT_ID
from bcipy.helpers import save
from bcipy.helpers.save import init_save_data_structure
from mockito import any, unstub, when


class TestSave(unittest.TestCase):
    """This is Test Case for Saving BCI data."""

    def setUp(self):
        # set up the needed paths and initial data save structure

        self.data_save_path = tempfile.mkdtemp()
        self.user_information = 'test_user_002'
        self.parameters_used = DEFAULT_PARAMETERS_PATH
        self.task = 'RSVP Calibration'
        self.experiment = DEFAULT_EXPERIMENT_ID

        self.save_folder_name = init_save_data_structure(
            self.data_save_path,
            self.user_information,
            self.parameters_used,
            self.task)

        self.dt = '10'

        # mock save modules use of strftime to return an empty string
        when(save).strftime(any(), any()).thenReturn(self.dt)

    def tearDown(self):
        # clean up by removing the data folder we used for testing
        shutil.rmtree(self.data_save_path)
        unstub()

    def test_init_save_data_structure_creates_correct_save_folder(self):

        # assert the save folder was created
        self.assertTrue(os.path.isdir(self.save_folder_name))

    def test_parameter_file_copies(self):

        # construct the path of the parameters
        param_path = self.save_folder_name + f'/{DEFAULT_PARAMETER_FILENAME}'

        # assert that the params file was created in the correct location
        self.assertTrue(os.path.isfile(param_path))

    def test_save_structure_adds_default_experiment(self):
        self.assertIn(DEFAULT_EXPERIMENT_ID, self.save_folder_name)

    def test_save_structure_adds_experiment_id_when_provided_as_argument(self):
        experiment_id = 'test'
        response = init_save_data_structure(
            self.data_save_path,
            self.user_information,
            self.parameters_used,
            self.task,
            experiment_id=experiment_id)
        task = self.task.replace(' ', '_')
        expected = (
            f'{self.data_save_path}{experiment_id}'
            f'/{self.user_information}/{self.user_information}_{task}_{self.dt}'
        )

        self.assertEqual(response, expected)

    def test_throws_error_if_given_incorrect_params_path(self):

        # try passing a parameters file that does not exist
        with self.assertRaises(Exception):
            init_save_data_structure(
                self.data_save_path,
                self.user_information,
                'does_not_exist.json',
                self.task,
                self.experiment)


if __name__ == '__main__':
    unittest.main()



from typing import Tuple
import numpy as np
import pywt


def continuous_wavelet_transform(
        data: np.ndarray, freq: int, fs: int, wavelet="cmor1.5-1.0") -> Tuple[np.ndarray, float]:
    """
    Transform data into frequency domain using Continuous Wavelet Transform (CWT).
    Keeps only a single wavelet scale, specified by `freq`.

    Args:
        data (np.ndarray): shape (trials, channels, time)
        freq (int): frequency of wavelet to keep
        fs (int): sampling rate of data (Hz)
        wavelet (str): name of wavelet to use (see pywt.families() and pywt.wavelist())
            default: "cmor1.5-1.0" (complex morlet wavelet)
            https://pywavelets.readthedocs.io/en/latest/ref/cwt.html#complex-morlet-wavelets

    Returns:
        np.ndarray: frequency domain data, shape (trials, wavelet_scales*channels, time)
                    Note that here we only use 1 wavelet_scale.
        float: wavelet scale used
    """
    scales = pywt.central_frequency(wavelet) * fs / np.array(freq)
    all_coeffs = []
    for trial in data:
        coeffs, _ = pywt.cwt(trial, scales, wavelet)  # shape == (scales, channels, time)
        all_coeffs.append(coeffs)

    final_data = np.stack(all_coeffs)
    if np.any(np.iscomplex(final_data)):
        final_data = np.abs(final_data) ** 2

    # have shape == (trials, freqs, channels, time)
    # want shape == (trials, freqs*channels, time)
    return final_data.reshape(final_data.shape[0], -1, final_data.shape[-1]), scales
